@@ -1,40 +1,47 @@
+from __future__ import unicode_literals
+
 import base64
-import cookielib
-import httplib
 import logging
-import mimetools
 import mimetypes
 import os
+import random
 import shutil
-import urllib
-import urllib2
-from StringIO import StringIO
-from urlparse import urlparse, urlunparse
-
-try:
-    from cStringIO import StringIO
-except ImportError:
-    from StringIO import StringIO
-
-try:
-    # Specifically import json_loads, to work around some issues with
-    # installations containing incompatible modules named "json".
-    from json import loads as json_loads
-except ImportError:
-    from simplejson import loads as json_loads
-
-try:
-    # In python 2.6, parse_qsl was deprectated in cgi, and
-    # moved to urlparse.
-    from urlparse import parse_qsl
-except ImportError:
-    from cgi import parse_qsl
+import sys
+from io import BytesIO
+from json import loads as json_loads
 
+import six
+from six.moves.http_client import UNAUTHORIZED
+from six.moves.http_cookiejar import Cookie, CookieJar, MozillaCookieJar
+from six.moves.urllib.error import HTTPError, URLError
+from six.moves.urllib.parse import parse_qsl, urlencode, urlparse, urlunparse
+from six.moves.urllib.request import (
+    BaseHandler,
+    HTTPBasicAuthHandler,
+    HTTPCookieProcessor,
+    HTTPDigestAuthHandler,
+    HTTPErrorProcessor,
+    HTTPPasswordMgr,
+    ProxyHandler,
+    Request as URLRequest,
+    build_opener,
+    install_opener,
+    urlopen)
 
 from rbtools import get_package_version
+from rbtools.api.cache import APICache
 from rbtools.api.errors import APIError, create_api_error, ServerInterfaceError
 from rbtools.utils.filesystem import get_home_path
 
+# Python 2.7.9+ added strict HTTPS certificate validation (finally). These APIs
+# don't exist everywhere so soft-import them.
+try:
+    import ssl
+    from six.moves.urllib.request import HTTPSHandler
+except ImportError:
+    ssl = None
+    HTTPSHandler = None
+
 
 RBTOOLS_COOKIE_FILE = '.rbtools-cookies'
 RB_COOKIE_NAME = 'rbsessionid'
@@ -52,16 +59,14 @@ class HttpRequest(object):
         # key with dashes.
         query_args = dict([
             (key.replace('_', '-'), value)
-            for key, value in query_args.iteritems()
+            for key, value in six.iteritems(query_args)
         ])
 
         # Add the query arguments to the url
-        # TODO: Make work with Python < 2.6. In 2.6
-        # parse_qsl was moved from cgi to urlparse.
         url_parts = list(urlparse(url))
         query = dict(parse_qsl(url_parts[4]))
         query.update(query_args)
-        url_parts[4] = urllib.urlencode(query)
+        url_parts[4] = urlencode(query)
         self.url = urlunparse(url_parts)
 
     def add_field(self, name, value):
@@ -80,7 +85,7 @@ class HttpRequest(object):
         del self._files[filename]
 
     def encode_multipart_formdata(self):
-        """ Encodes data for use in an HTTP request.
+        """Encodes data for use in an HTTP request.
 
         Parameters:
             fields - the fields to be encoded.  This should be a dict in a
@@ -91,49 +96,80 @@ class HttpRequest(object):
         if not (self._fields or self._files):
             return None, None
 
-        NEWLINE = '\r\n'
-        BOUNDARY = mimetools.choose_boundary()
-        content = StringIO()
+        NEWLINE = b'\r\n'
+        BOUNDARY = self._make_mime_boundary()
+        content = BytesIO()
 
         for key in self._fields:
-            content.write('--' + BOUNDARY + NEWLINE)
-            content.write('Content-Disposition: form-data; name="%s"' % key)
+            content.write(b'--' + BOUNDARY + NEWLINE)
+            content.write(b'Content-Disposition: form-data; '
+                          b'name="%s"' % key.encode('utf-8'))
             content.write(NEWLINE + NEWLINE)
-            content.write(str(self._fields[key]) + NEWLINE)
+
+            if isinstance(self._fields[key], six.binary_type):
+                content.write(self._fields[key] + NEWLINE)
+            else:
+                content.write(six.text_type(self._fields[key]).encode('utf-8')
+                              + NEWLINE)
 
         for key in self._files:
             filename = self._files[key]['filename']
             value = self._files[key]['content']
-            mime_type = (mimetypes.guess_type(filename)[0] or
-                         'application/octet-stream')
-            content.write('--' + BOUNDARY + NEWLINE)
-            content.write('Content-Disposition: form-data; name="%s"; ' % key)
-            content.write('filename="%s"' % filename + NEWLINE)
-            content.write('Content-Type: %s' % mime_type + NEWLINE)
-            content.write(NEWLINE)
-            content.write(value)
+
+            mime_type = mimetypes.guess_type(filename)[0]
+            if mime_type:
+                mime_type = mime_type.encode('utf-8')
+            else:
+                mime_type = b'application/octet-stream'
+
+            content.write(b'--' + BOUNDARY + NEWLINE)
+            content.write(b'Content-Disposition: form-data; name="%s"; '
+                          % key.encode('utf-8'))
+            content.write(b'filename="%s"' % filename.encode('utf-8')
+                          + NEWLINE)
+            content.write(b'Content-Type: %s' % mime_type + NEWLINE)
             content.write(NEWLINE)
 
-        content.write('--' + BOUNDARY + '--' + NEWLINE + NEWLINE)
-        content_type = 'multipart/form-data; boundary=%s' % BOUNDARY
+            if isinstance(value, six.text_type):
+                content.write(value.encode('utf-8'))
+            else:
+                content.write(value)
+
+            content.write(NEWLINE)
+
+        content.write(b'--' + BOUNDARY + b'--' + NEWLINE + NEWLINE)
+        content_type = b'multipart/form-data; boundary=%s' % BOUNDARY
 
         return content_type, content.getvalue()
 
+    def _make_mime_boundary(self):
+        """Create a mime boundary.
 
-class Request(urllib2.Request):
+        This exists because mimetools.choose_boundary() is gone in Python 3.x,
+        and email.generator._make_boundary isn't really appropriate to use
+        here.
+        """
+        fmt = '%%0%dd' % len(repr(sys.maxsize - 1))
+        token = random.randrange(sys.maxsize)
+        return (b'=' * 15) + (fmt % token).encode('utf-8') + b'=='
+
+
+class Request(URLRequest):
     """A request which contains a method attribute."""
-    def __init__(self, url, body='', headers={}, method="PUT"):
-        urllib2.Request.__init__(self, url, body, headers)
+    def __init__(self, url, body='', headers={}, method='PUT'):
+        URLRequest.__init__(self, url, body, headers)
         self.method = method
 
     def get_method(self):
         return self.method
 
 
-class PresetHTTPAuthHandler(urllib2.BaseHandler):
-    """urllib2 handler that presets the use of HTTP Basic Auth."""
+class PresetHTTPAuthHandler(BaseHandler):
+    """Handler that presets the use of HTTP Basic Auth."""
     handler_order = 480  # After Basic auth
 
+    AUTH_HEADER = 'Authorization'
+
     def __init__(self, url, password_mgr):
         self.url = url
         self.password_mgr = password_mgr
@@ -145,31 +181,35 @@ class PresetHTTPAuthHandler(urllib2.BaseHandler):
         self.used = False
 
     def http_request(self, request):
-        if not self.used and self.password_mgr.rb_user:
-            # Note that we call password_mgr.find_user_password to get the
-            # username and password we're working with.
-            username, password = \
-                self.password_mgr.find_user_password('Web API', self.url)
-            raw = '%s:%s' % (username, password)
-            request.add_header(
-                urllib2.HTTPBasicAuthHandler.auth_header,
-                'Basic %s' % base64.b64encode(raw).strip())
-            self.used = True
+        if not self.used:
+            if self.password_mgr.api_token:
+                request.add_header(self.AUTH_HEADER,
+                                   'token %s' % self.password_mgr.api_token)
+                self.used = True
+            elif self.password_mgr.rb_user:
+                # Note that we call password_mgr.find_user_password to get the
+                # username and password we're working with.
+                username, password = \
+                    self.password_mgr.find_user_password('Web API', self.url)
+                raw = '%s:%s' % (username, password)
+                request.add_header(self.AUTH_HEADER,
+                                   'Basic %s' % base64.b64encode(raw).strip())
+                self.used = True
 
         return request
 
     https_request = http_request
 
 
-class ReviewBoardHTTPErrorProcessor(urllib2.HTTPErrorProcessor):
+class ReviewBoardHTTPErrorProcessor(HTTPErrorProcessor):
     """Processes HTTP error codes.
 
-    Python 2.6 gets HTTP error code processing right, but 2.4 and 2.5
-    only accepts HTTP 200 and 206 as success codes. This handler
-    ensures that anything in the 200 range is a success.
+    Python 2.6 gets HTTP error code processing right, but 2.4 and 2.5 only
+    accepts HTTP 200 and 206 as success codes. This handler ensures that
+    anything in the 200 range, as well as 304, is a success.
     """
     def http_response(self, request, response):
-        if not (200 <= response.code < 300):
+        if not (200 <= response.code < 300 or response.code == 304):
             response = self.parent.error('http', request, response,
                                          response.code, response.msg,
                                          response.info())
@@ -179,13 +219,13 @@ class ReviewBoardHTTPErrorProcessor(urllib2.HTTPErrorProcessor):
     https_response = http_response
 
 
-class ReviewBoardHTTPBasicAuthHandler(urllib2.HTTPBasicAuthHandler):
+class ReviewBoardHTTPBasicAuthHandler(HTTPBasicAuthHandler):
     """Custom Basic Auth handler that doesn't retry excessively.
 
-    urllib2's HTTPBasicAuthHandler retries over and over, which is
-    useless. This subclass only retries once to make sure we've
-    attempted with a valid username and password. It will then fail so
-    we can use our own retry handler.
+    urllib's HTTPBasicAuthHandler retries over and over, which is useless. This
+    subclass only retries once to make sure we've attempted with a valid
+    username and password. It will then fail so we can use our own retry
+    handler.
 
     This also supports two-factor auth, for Review Board servers that
     support it. When requested by the server, the client will be prompted
@@ -197,7 +237,7 @@ class ReviewBoardHTTPBasicAuthHandler(urllib2.HTTPBasicAuthHandler):
     MAX_OTP_TOKEN_ATTEMPTS = 5
 
     def __init__(self, *args, **kwargs):
-        urllib2.HTTPBasicAuthHandler.__init__(self, *args, **kwargs)
+        HTTPBasicAuthHandler.__init__(self, *args, **kwargs)
         self._retried = False
         self._lasturl = ""
         self._needs_otp_token = False
@@ -216,7 +256,7 @@ class ReviewBoardHTTPBasicAuthHandler(urllib2.HTTPBasicAuthHandler):
 
         response = self._do_http_basic_auth(host, request, realm)
 
-        if response and response.code != httplib.UNAUTHORIZED:
+        if response and response.code != UNAUTHORIZED:
             self._retried = False
 
         return response
@@ -227,7 +267,7 @@ class ReviewBoardHTTPBasicAuthHandler(urllib2.HTTPBasicAuthHandler):
         if password is None:
             return None
 
-        raw = "%s:%s" % (user, password)
+        raw = '%s:%s' % (user, password)
         auth = 'Basic %s' % base64.b64encode(raw).strip()
 
         if (request.headers.get(self.auth_header, None) == auth and
@@ -242,7 +282,7 @@ class ReviewBoardHTTPBasicAuthHandler(urllib2.HTTPBasicAuthHandler):
         try:
             response = self.parent.open(request, timeout=request.timeout)
             return response
-        except urllib2.HTTPError, e:
+        except HTTPError as e:
             if e.code == 401:
                 headers = e.info()
                 otp_header = headers.get(self.OTP_TOKEN_HEADER, '')
@@ -272,7 +312,7 @@ class ReviewBoardHTTPBasicAuthHandler(urllib2.HTTPBasicAuthHandler):
         return None
 
 
-class ReviewBoardHTTPPasswordMgr(urllib2.HTTPPasswordMgr):
+class ReviewBoardHTTPPasswordMgr(HTTPPasswordMgr):
     """Adds HTTP authentication support for URLs.
 
     Python 2.4's password manager has a bug in http authentication
@@ -282,12 +322,13 @@ class ReviewBoardHTTPPasswordMgr(urllib2.HTTPPasswordMgr):
     See: http://bugs.python.org/issue974757
     """
     def __init__(self, reviewboard_url, rb_user=None, rb_pass=None,
-                 auth_callback=None, otp_token_callback=None):
-        urllib2.HTTPPasswordMgr.__init__(self)
+                 api_token=None, auth_callback=None, otp_token_callback=None):
+        HTTPPasswordMgr.__init__(self)
         self.passwd = {}
         self.rb_url = reviewboard_url
         self.rb_user = rb_user
         self.rb_pass = rb_pass
+        self.api_token = api_token
         self.auth_callback = auth_callback
         self.otp_token_callback = otp_token_callback
 
@@ -304,7 +345,7 @@ class ReviewBoardHTTPPasswordMgr(urllib2.HTTPPasswordMgr):
         else:
             # If this is an auth request for some other domain (since HTTP
             # handlers are global), fall back to standard password management.
-            return urllib2.HTTPPasswordMgr.find_user_password(self, realm, uri)
+            return HTTPPasswordMgr.find_user_password(self, realm, uri)
 
     def get_otp_token(self, uri, method):
         if self.otp_token_callback:
@@ -332,20 +373,20 @@ def create_cookie_jar(cookie_file=None):
             os.path.isfile(post_review_cookies)):
                 try:
                     shutil.copyfile(post_review_cookies, cookie_file)
-                    os.chmod(cookie_file, 0600)
-                except IOError, e:
+                    os.chmod(cookie_file, 0o600)
+                except IOError as e:
                     logging.warning("There was an error while copying "
-                                    "post-review's cookies: %s" % e)
+                                    "post-review's cookies: %s", e)
 
     if not os.path.isfile(cookie_file):
         try:
             open(cookie_file, 'w').close()
-            os.chmod(cookie_file, 0600)
-        except IOError, e:
-            logging.warning("There was an error while creating a "
-                            "cookie file: %s" % e)
+            os.chmod(cookie_file, 0o600)
+        except IOError as e:
+            logging.warning('There was an error while creating a '
+                            'cookie file: %s', e)
 
-    return cookielib.MozillaCookieJar(cookie_file), cookie_file
+    return MozillaCookieJar(cookie_file), cookie_file
 
 
 class ReviewBoardServer(object):
@@ -361,37 +402,45 @@ class ReviewBoardServer(object):
     for their credentials using this mechanism.
     """
     def __init__(self, url, cookie_file=None, username=None, password=None,
-                 agent=None, session=None, disable_proxy=False,
-                 auth_callback=None, otp_token_callback=None):
-        self.url = url
-        if self.url[-1] != '/':
-            self.url += '/'
+                 api_token=None, agent=None, session=None, disable_proxy=False,
+                 auth_callback=None, otp_token_callback=None,
+                 verify_ssl=True, save_cookies=True):
+        if not url.endswith('/'):
+            url += '/'
 
-        self.url = self.url + 'api/'
-        self.cookie_jar, self.cookie_file = create_cookie_jar(
-            cookie_file=cookie_file)
+        self.url = url + 'api/'
 
-        try:
-            self.cookie_jar.load(ignore_expires=True)
-        except IOError:
-            pass
+        self.save_cookies = save_cookies
+
+        if self.save_cookies:
+            self.cookie_jar, self.cookie_file = create_cookie_jar(
+                cookie_file=cookie_file)
+
+            try:
+                self.cookie_jar.load(ignore_expires=True)
+            except IOError:
+                pass
+        else:
+            self.cookie_jar = CookieJar()
+            self.cookie_file = None
+
+        # Get the cookie domain from the url. If the domain
+        # does not contain a '.' (e.g. 'localhost'), we assume
+        # it is a local domain and suffix it (See RFC 2109).
+        parsed_url = urlparse(url)
+        self.domain = parsed_url[1].partition(':')[0]  # Remove Port.
+
+        if self.domain.count('.') < 1:
+            self.domain = '%s.local' % self.domain
 
         if session:
-            parsed_url = urlparse(url)
-            # Get the cookie domain from the url. If the domain
-            # does not contain a '.' (e.g. 'localhost'), we assume
-            # it is a local domain and suffix it (See RFC 2109).
-            domain = parsed_url[1].partition(':')[0]  # Remove Port.
-            if domain.count('.') < 1:
-                domain = "%s.local" % domain
-
-            cookie = cookielib.Cookie(
+            cookie = Cookie(
                 version=0,
                 name=RB_COOKIE_NAME,
                 value=session,
                 port=None,
                 port_specified=False,
-                domain=domain,
+                domain=self.domain,
                 domain_specified=True,
                 domain_initial_dot=True,
                 path=parsed_url[2],
@@ -403,12 +452,25 @@ class ReviewBoardServer(object):
                 comment_url=None,
                 rest={'HttpOnly': None})
             self.cookie_jar.set_cookie(cookie)
-            self.cookie_jar.save()
+
+            if self.save_cookies:
+                self.cookie_jar.save()
+
+        if username:
+            # If the username parameter is given, we have to clear the session
+            # cookie manually or it will override the username:password
+            # combination retrieved from the authentication callback.
+            try:
+                self.cookie_jar.clear(self.domain, parsed_url[2],
+                                      RB_COOKIE_NAME)
+            except KeyError:
+                pass
 
         # Set up the HTTP libraries to support all of the features we need.
         password_mgr = ReviewBoardHTTPPasswordMgr(self.url,
                                                   username,
                                                   password,
+                                                  api_token,
                                                   auth_callback,
                                                   otp_token_callback)
         self.preset_auth_handler = PresetHTTPAuthHandler(self.url,
@@ -416,13 +478,17 @@ class ReviewBoardServer(object):
 
         handlers = []
 
+        if not verify_ssl:
+            context = ssl._create_unverified_context()
+            handlers.append(HTTPSHandler(context=context))
+
         if disable_proxy:
-            handlers.append(urllib2.ProxyHandler({}))
+            handlers.append(ProxyHandler({}))
 
         handlers += [
-            urllib2.HTTPCookieProcessor(self.cookie_jar),
+            HTTPCookieProcessor(self.cookie_jar),
             ReviewBoardHTTPBasicAuthHandler(password_mgr),
-            urllib2.HTTPDigestAuthHandler(password_mgr),
+            HTTPDigestAuthHandler(password_mgr),
             self.preset_auth_handler,
             ReviewBoardHTTPErrorProcessor(),
         ]
@@ -430,18 +496,45 @@ class ReviewBoardServer(object):
         if agent:
             self.agent = agent
         else:
-            self.agent = 'RBTools/' + get_package_version()
+            self.agent = ('RBTools/' + get_package_version()).encode('utf-8')
 
-        opener = urllib2.build_opener(*handlers)
+        opener = build_opener(*handlers)
         opener.addheaders = [
-            ('User-agent', self.agent),
+            (b'User-agent', self.agent),
         ]
-        urllib2.install_opener(opener)
+        install_opener(opener)
+
+        self._cache = None
+        self._urlopen = urlopen
+
+    def enable_cache(self, cache_location=None, in_memory=False):
+        """Enable caching for all future HTTP requests.
+
+        The cache will be created at the default location if none is provided.
+
+        If the in_memory parameter is True, the cache will be created in memory
+        instead of on disk. This overrides the cache_location parameter.
+        """
+        if not self._cache:
+            self._cache = APICache(create_db_in_memory=in_memory,
+                                   db_location=cache_location)
+
+            self._urlopen = self._cache.make_request
 
     def login(self, username, password):
         """Reset the user information"""
         self.preset_auth_handler.reset(username, password)
 
+    def logout(self):
+        """Logs the user out of the session."""
+        self.preset_auth_handler.reset(None, None)
+        self.make_request(HttpRequest('%ssession/' % self.url,
+                                      method='DELETE'))
+        self.cookie_jar.clear(self.domain)
+
+        if self.save_cookies:
+            self.cookie_jar.save()
+
     def process_error(self, http_status, data):
         """Processes an error, raising an APIError with the information."""
         try:
@@ -471,23 +564,24 @@ class ReviewBoardServer(object):
 
             if body:
                 headers.update({
-                    'Content-Type': content_type,
-                    'Content-Length': str(len(body)),
+                    b'Content-Type': content_type,
+                    b'Content-Length': str(len(body)),
                 })
             else:
-                headers['Content-Length'] = "0"
+                headers[b'Content-Length'] = '0'
 
             r = Request(request.url.encode('utf-8'), body, headers,
-                        request.method)
-            rsp = urllib2.urlopen(r)
-        except urllib2.HTTPError, e:
+                        request.method.encode('utf-8'))
+            rsp = self._urlopen(r)
+        except HTTPError as e:
             self.process_error(e.code, e.read())
-        except urllib2.URLError, e:
-            raise ServerInterfaceError("%s" % e.reason)
+        except URLError as e:
+            raise ServerInterfaceError('%s' % e.reason)
 
-        try:
-            self.cookie_jar.save()
-        except IOError:
-            pass
+        if self.save_cookies:
+            try:
+                self.cookie_jar.save()
+            except IOError:
+                pass
 
         return rsp
