@@ -3,13 +3,16 @@ import os
 import re
 import sys
 
-from rbtools.clients import SCMClient, RepositoryInfo
-from rbtools.clients.errors import (InvalidRevisionSpecError,
+from rbtools.clients import PatchResult, SCMClient, RepositoryInfo
+from rbtools.clients.errors import (AmendError, MergeError, PushError,
+                                    InvalidRevisionSpecError,
                                     TooManyRevisionsError)
 from rbtools.clients.perforce import PerforceClient
 from rbtools.clients.svn import SVNClient, SVNRepositoryInfo
-from rbtools.utils.checks import check_install
+from rbtools.utils.checks import check_install, is_valid_version
 from rbtools.utils.console import edit_text
+from rbtools.utils.diffs import (normalize_patterns,
+                                 remove_filenames_matching_patterns)
 from rbtools.utils.process import die, execute
 
 
@@ -21,12 +24,37 @@ class GitClient(SCMClient):
     """
     name = 'Git'
 
+    supports_diff_exclude_patterns = True
+    supports_patch_revert = True
+
+    can_amend_commit = True
+    can_merge = True
+    can_push_upstream = True
+    can_delete_branch = True
+
     def __init__(self, **kwargs):
         super(GitClient, self).__init__(**kwargs)
         # Store the 'correct' way to invoke git, just plain old 'git' by
         # default.
         self.git = 'git'
 
+        self._original_cwd = None
+        self._git_version = None
+
+        version_str = execute([self.git, 'version'], ignore_errors=True,
+                              none_on_ignored_error=True)
+
+        if version_str:
+            m = re.search(r'(\d+).(\d+).(\d+)', version_str)
+
+            if m:
+                self._git_version = (int(m.group(1)),
+                                     int(m.group(2)),
+                                     int(m.group(3)))
+
+        self._supports_git_config_flag = (self._git_version and
+                                          self._git_version >= (1, 8, 0))
+
     def parse_revision_spec(self, revisions=[]):
         """Parses the given revision spec.
 
@@ -140,6 +168,11 @@ class GitClient(SCMClient):
         return result
 
     def get_repository_info(self):
+        """Get repository information for the current Git working tree.
+
+        This function changes the directory to the top level directory of the
+        current working tree.
+        """
         if not check_install(['git', '--help']):
             # CreateProcess (launched via subprocess, used by check_install)
             # does not automatically append .cmd for things it finds in PATH.
@@ -164,6 +197,13 @@ class GitClient(SCMClient):
                        ignore_errors=True).strip()
         self.bare = bare in ('true', '1')
 
+        # If we are not working in a bare repository, then we will change
+        # directory to the top level working tree lose our original position.
+        # However, we need the original working directory for file exclusion
+        # patterns, so we save it here.
+        if self._original_cwd is None:
+            self._original_cwd = os.getcwd()
+
         # Running in directories other than the top level of
         # of a work-tree would result in broken diffs on the server
         if not self.bare:
@@ -241,10 +281,10 @@ class GitClient(SCMClient):
                     ignore_errors=True)
 
                 if (version_parts and svn_remote and
-                    not self.is_valid_version((int(version_parts.group(1)),
-                                               int(version_parts.group(2)),
-                                               int(version_parts.group(3))),
-                                              (1, 5, 4))):
+                    not is_valid_version((int(version_parts.group(1)),
+                                          int(version_parts.group(2)),
+                                          int(version_parts.group(3))),
+                                         (1, 5, 4))):
                     die("Your installation of git-svn must be upgraded to "
                         "version 1.5.4 or later")
 
@@ -309,7 +349,6 @@ class GitClient(SCMClient):
             self.type = "git"
             return RepositoryInfo(path=url, base_path='',
                                   supports_parent_diffs=True)
-
         return None
 
     def _strip_heads_prefix(self, ref):
@@ -330,18 +369,6 @@ class GitClient(SCMClient):
             ignore_errors=True).rstrip("\n")
         return (upstream_branch, origin_url)
 
-    def is_valid_version(self, actual, expected):
-        """
-        Takes two tuples, both in the form:
-            (major_version, minor_version, micro_version)
-        Returns true if the actual version is greater than or equal to
-        the expected version, and false otherwise.
-        """
-        return ((actual[0] > expected[0]) or
-                (actual[0] == expected[0] and actual[1] > expected[1]) or
-                (actual[0] == expected[0] and actual[1] == expected[1] and
-                 actual[2] >= expected[2]))
-
     def scan_for_server(self, repository_info):
         # Scan first for dot files, since it's faster and will cover the
         # user's $HOME/.reviewboardrc
@@ -380,10 +407,10 @@ class GitClient(SCMClient):
 
     def get_parent_branch(self):
         """Returns the parent branch."""
+        parent_branch = getattr(self.options, 'parent_branch', None)
+
         if self.type == 'perforce':
-            parent_branch = self.options.parent_branch or 'p4'
-        else:
-            parent_branch = self.options.parent_branch
+            parent_branch = parent_branch or 'p4'
 
         return parent_branch
 
@@ -407,7 +434,8 @@ class GitClient(SCMClient):
 
         return execute([self.git, 'rev-parse'] + revisions).strip().split('\n')
 
-    def diff(self, revisions, files=[], extra_args=[]):
+    def diff(self, revisions, include_files=[], exclude_patterns=[],
+             extra_args=[]):
         """Perform a diff using the given revisions.
 
         If no revisions are specified, this will do a diff of the contents of
@@ -420,6 +448,10 @@ class GitClient(SCMClient):
         make sense given the requested revisions and the tracking branch, this
         will also return a parent diff.
         """
+        exclude_patterns = normalize_patterns(exclude_patterns,
+                                              self._get_root_directory(),
+                                              cwd=self.original_cwd)
+
         try:
             merge_base = revisions['parent_base']
         except KeyError:
@@ -428,13 +460,16 @@ class GitClient(SCMClient):
         diff_lines = self.make_diff(merge_base,
                                     revisions['base'],
                                     revisions['tip'],
-                                    files)
+                                    include_files,
+                                    exclude_patterns)
 
         if 'parent_base' in revisions:
             parent_diff_lines = self.make_diff(merge_base,
                                                revisions['parent_base'],
                                                revisions['base'],
-                                               files)
+                                               include_files,
+                                               exclude_patterns)
+
             base_commit_id = revisions['parent_base']
         else:
             parent_diff_lines = None
@@ -447,47 +482,105 @@ class GitClient(SCMClient):
             'base_commit_id': base_commit_id,
         }
 
-    def make_diff(self, merge_base, base, tip, files):
+    def make_diff(self, merge_base, base, tip, include_files,
+                  exclude_patterns):
         """Performs a diff on a particular branch range."""
         rev_range = "%s..%s" % (base, tip)
 
-        if files:
-            files = ['--'] + files
+        if include_files:
+            include_files = ['--'] + include_files
+
+        git_cmd = [self.git]
+
+        if self._supports_git_config_flag:
+            git_cmd.extend(['-c', 'core.quotepath=false'])
 
         if self.type in ('svn', 'perforce'):
-            diff_cmd = [self.git, 'diff', '--no-color', '--no-prefix', '-r',
-                        '-u', rev_range]
-        elif self.type == "git":
-            diff_cmd = [self.git, 'diff', '--no-color', '--full-index',
-                        '--ignore-submodules', rev_range]
+            diff_cmd_params = ['--no-color', '--no-prefix', '-r', '-u']
+        elif self.type == 'git':
+            diff_cmd_params = ['--no-color', '--full-index',
+                               '--ignore-submodules']
+
+            if self._supports_git_config_flag:
+                git_cmd.extend(['-c', 'diff.noprefix=false'])
 
             if (self.capabilities is not None and
                 self.capabilities.has_capability('diffs', 'moved_files')):
-                diff_cmd.append('-M')
+                diff_cmd_params.append('-M')
             else:
-                diff_cmd.append('--no-renames')
+                diff_cmd_params.append('--no-renames')
         else:
-            return None
+            assert False
 
         # By default, don't allow using external diff commands. This prevents
         # things from breaking horribly if someone configures a graphical diff
         # viewer like p4merge or kaleidoscope. This can be overridden by
         # setting GIT_USE_EXT_DIFF = True in ~/.reviewboardrc
-        if self.user_config.get('GIT_USE_EXT_DIFF', False):
-            diff_cmd.append('--no-ext-diff')
+        if not self.config.get('GIT_USE_EXT_DIFF', False):
+            diff_cmd_params.append('--no-ext-diff')
 
-        diff_lines = execute(diff_cmd + files,
-                             split_lines=True,
-                             with_errors=False,
-                             ignore_errors=True,
-                             none_on_ignored_error=True)
+        diff_cmd = git_cmd + ['diff'] + diff_cmd_params
+
+        if exclude_patterns:
+            # If we have specified files to exclude, we will get a list of all
+            # changed files and run `git diff` on each un-excluded file
+            # individually.
+            changed_files_cmd = git_cmd + ['diff-tree'] + diff_cmd_params
+            if self.type == 'git':
+                changed_files_cmd.append('-r')
+
+            changed_files = execute(
+                changed_files_cmd + [rev_range] + include_files,
+                split_lines=True,
+                with_errors=False,
+                ignore_errors=True,
+                none_on_ignored_error=True,
+                log_output_on_error=False)
+
+            # The output of git diff-tree will be a list of entries that have
+            # changed between the two revisions that we give it. The last part
+            # of the line is the name of the file that has changed.
+            changed_files = remove_filenames_matching_patterns(
+                (filename.split()[-1] for filename in changed_files),
+                exclude_patterns, base_dir=self._get_root_directory())
+
+            diff_lines = []
+
+            for filename in changed_files:
+                lines = execute(diff_cmd + [rev_range, '--', filename],
+                                split_lines=True,
+                                with_errors=False,
+                                ignore_errors=True,
+                                none_on_ignored_error=True,
+                                log_output_on_error=False,
+                                results_unicode=False)
+
+                if lines is None:
+                    logging.error(
+                        'Could not get diff for all files (git-diff failed '
+                        'for "%s"). Refusing to return a partial diff.' %
+                        filename)
+
+                    diff_lines = None
+                    break
+
+                diff_lines += lines
+
+        else:
+            diff_lines = execute(diff_cmd + [rev_range] + include_files,
+                                 split_lines=True,
+                                 with_errors=False,
+                                 ignore_errors=True,
+                                 none_on_ignored_error=True,
+                                 log_output_on_error=False,
+                                 results_unicode=False)
 
         if self.type == 'svn':
             return self.make_svn_diff(merge_base, diff_lines)
         elif self.type == 'perforce':
             return self.make_perforce_diff(merge_base, diff_lines)
         else:
-            return ''.join(diff_lines)
+            return b''.join(diff_lines)
 
     def make_svn_diff(self, merge_base, diff_lines):
         """
@@ -500,46 +593,49 @@ class GitClient(SCMClient):
         if not rev:
             return None
 
-        diff_data = ""
-        filename = ""
+        diff_data = b''
+        original_file = b''
+        filename = b''
         newfile = False
 
-        for line in diff_lines:
-            if line.startswith("diff "):
+        for i, line in enumerate(diff_lines):
+            if line.startswith(b'diff '):
                 # Grab the filename and then filter this out.
                 # This will be in the format of:
                 #
                 # diff --git a/path/to/file b/path/to/file
-                info = line.split(" ")
-                diff_data += "Index: %s\n" % info[2]
-                diff_data += "=" * 67
-                diff_data += "\n"
-            elif line.startswith("index "):
+                info = line.split(b' ')
+                diff_data += b'Index: %s\n' % info[2]
+                diff_data += b'=' * 67
+                diff_data += b'\n'
+            elif line.startswith(b'index '):
                 # Filter this out.
                 pass
-            elif line.strip() == "--- /dev/null":
+            elif line.strip() == b'--- /dev/null':
                 # New file
                 newfile = True
-            elif line.startswith("--- "):
+            elif (line.startswith(b'--- ') and i + 1 < len(diff_lines) and
+                  diff_lines[i + 1].startswith(b'+++ ')):
                 newfile = False
-                diff_data += "--- %s\t(revision %s)\n" % \
-                             (line[4:].strip(), rev)
-            elif line.startswith("+++ "):
+                original_file = line[4:].strip()
+                diff_data += b'--- %s\t(revision %s)\n' % (original_file, rev)
+            elif line.startswith(b'+++ '):
                 filename = line[4:].strip()
                 if newfile:
-                    diff_data += "--- %s\t(revision 0)\n" % filename
-                    diff_data += "+++ %s\t(revision 0)\n" % filename
+                    diff_data += b'--- %s\t(revision 0)\n' % filename
+                    diff_data += b'+++ %s\t(revision 0)\n' % filename
                 else:
                     # We already printed the "--- " line.
-                    diff_data += "+++ %s\t(working copy)\n" % filename
-            elif line.startswith("new file mode"):
+                    diff_data += b'+++ %s\t(working copy)\n' % original_file
+            elif (line.startswith(b'new file mode') or
+                  line.startswith(b'deleted file mode')):
                 # Filter this out.
                 pass
-            elif line.startswith("Binary files "):
+            elif line.startswith(b'Binary files '):
                 # Add the following so that we know binary files were
                 # added/changed.
-                diff_data += "Cannot display: file marked as a binary type.\n"
-                diff_data += "svn:mime-type = application/octet-stream\n"
+                diff_data += b'Cannot display: file marked as a binary type.\n'
+                diff_data += b'svn:mime-type = application/octet-stream\n'
             else:
                 diff_data += line
 
@@ -547,15 +643,16 @@ class GitClient(SCMClient):
 
     def make_perforce_diff(self, merge_base, diff_lines):
         """Format the output of git diff to look more like perforce's."""
-        diff_data = ''
-        filename = ''
-        p4rev = ''
+        diff_data = b''
+        filename = b''
+        p4rev = b''
 
         # Find which depot changelist we're based on
         log = execute([self.git, 'log', merge_base], ignore_errors=True)
 
         for line in log:
-            m = re.search(r'[rd]epo.-paths = "(.+)": change = (\d+).*\]', log, re.M)
+            m = re.search(br'[rd]epo.-paths = "(.+)": change = (\d+).*\]',
+                          log, re.M)
 
             if m:
                 base_path = m.group(1).strip()
@@ -565,35 +662,36 @@ class GitClient(SCMClient):
                 # We should really raise an error here, base_path is required
                 pass
 
-        for line in diff_lines:
-            if line.startswith('diff '):
+        for i, line in enumerate(diff_lines):
+            if line.startswith(b'diff '):
                 # Grab the filename and then filter this out.
                 # This will be in the format of:
                 #    diff --git a/path/to/file b/path/to/file
-                filename = line.split(' ')[2].strip()
-            elif (line.startswith('index ') or
-                  line.startswith('new file mode ')):
+                filename = line.split(b' ')[2].strip()
+            elif (line.startswith(b'index ') or
+                  line.startswith(b'new file mode ')):
                 # Filter this out
                 pass
-            elif line.startswith('--- '):
+            elif (line.startswith(b'--- ') and i + 1 < len(diff_lines) and
+                  diff_lines[i + 1].startswith(b'+++ ')):
                 data = execute(
                     ['p4', 'files', base_path + filename + '@' + p4rev],
-                    ignore_errors=True)
-                m = re.search(r'^%s%s#(\d+).*$' % (re.escape(base_path),
-                                                   re.escape(filename)),
+                    ignore_errors=True, results_unicode=False)
+                m = re.search(br'^%s%s#(\d+).*$' % (re.escape(base_path),
+                                                    re.escape(filename)),
                               data, re.M)
                 if m:
-                    fileVersion = m.group(1).strip()
+                    file_version = m.group(1).strip()
                 else:
-                    fileVersion = 1
+                    file_version = 1
 
-                diff_data += '--- %s%s\t%s%s#%s\n' % (base_path, filename,
-                                                      base_path, filename,
-                                                      fileVersion)
-            elif line.startswith('+++ '):
+                diff_data += b'--- %s%s\t%s%s#%s\n' % (base_path, filename,
+                                                       base_path, filename,
+                                                       file_version)
+            elif line.startswith(b'+++ '):
                 # TODO: add a real timestamp
-                diff_data += '+++ %s%s\t%s\n' % (base_path, filename,
-                                                 'TIMESTAMP')
+                diff_data += b'+++ %s%s\t%s\n' % (base_path, filename,
+                                                  b'TIMESTAMP')
             else:
                 diff_data += line
 
@@ -609,21 +707,71 @@ class GitClient(SCMClient):
                           '--untracked-files=no'])
         return status != ''
 
-    def apply_patch(self, patch_file, base_path=None, base_dir=None, p=None):
+    def amend_commit_description(self, message, revisions):
+        """Update a commit message to the given string.
+
+        Since git can amend only the most recent commit, an AmendError will be
+        raised if revisions points to a commit other than the the most recent
+        commit.
+        """
+        if revisions and revisions['tip']:
+            commit_ids = execute([self.git, 'rev-parse', 'HEAD',
+                                  revisions['tip']], split_lines=True)
+            head_id = commit_ids[0].strip()
+            revision_id = commit_ids[1].strip()
+
+            if head_id != revision_id:
+                raise AmendError('Commit "%s" is not the latest commit, '
+                                 'and thus cannot be modified' % revision_id)
+
+        execute([self.git, 'commit', '--amend', '-m', message])
+
+    def apply_patch(self, patch_file, base_path=None, base_dir=None, p=None,
+                    revert=False):
         """Apply the given patch to index.
 
         This will take the given patch file and apply it to the index,
         scheduling all changes for commit.
         """
+        cmd = ['git', 'apply', '-3']
+
+        if revert:
+            cmd.append('-R')
+
         if p:
-            cmd = ['git', 'apply', '--index', '-p', p, patch_file]
+            cmd += ['-p', p]
+
+        cmd.append(patch_file)
+
+        rc, data = self._execute(cmd, with_errors=True, return_error_code=True)
+
+        if rc == 0:
+            return PatchResult(applied=True, patch_output=data)
+        elif 'with conflicts' in data:
+            return PatchResult(
+                applied=True,
+                has_conflicts=True,
+                conflicting_files=[
+                    line.split(' ', 1)[1]
+                    for line in data.splitlines()
+                    if line.startswith('U')
+                ],
+                patch_output=data)
         else:
-            cmd = ['git', 'apply', '--index', patch_file]
+            return PatchResult(applied=False, patch_output=data)
 
-        self._execute(cmd)
+    def create_commit(self, message, author, run_editor,
+                      files=[], all_files=False):
+        """Commits the given modified files.
 
-    def create_commit(self, message, author, files=[], all_files=False):
-        modified_message = edit_text(message)
+        This is expected to be called after applying a patch. This commits the
+        patch using information from the review request, opening the commit
+        message in $EDITOR to allow the user to update it.
+        """
+        if run_editor:
+            modified_message = edit_text(message)
+        else:
+            modified_message = message
 
         if all_files:
             execute(['git', 'add', '--all', ':/'])
@@ -633,7 +781,89 @@ class GitClient(SCMClient):
         execute(['git', 'commit', '-m', modified_message,
                  '--author="%s <%s>"' % (author.fullname, author.email)])
 
+    def delete_branch(self, branch_name, merged_only=True):
+        """Deletes the specified branch.
+
+        If merged_only is False, then the branch will be deleted even if not
+        yet merged into an upstream branch.
+        """
+        if merged_only:
+            delete_flag = '-d'
+        else:
+            delete_flag = '-D'
+
+        execute(['git', 'branch', delete_flag, branch_name])
+
+    def merge(self, target, destination, message, author, squash=False,
+              run_editor=False):
+        """Merges the target branch with destination branch."""
+        rc, output = execute(
+            ['git', 'checkout', destination],
+            ignore_errors=True,
+            return_error_code=True)
+
+        if rc:
+            raise MergeError("Could not checkout to branch '%s'.\n\n%s" %
+                             (destination, output))
+
+        if squash:
+            method = '--squash'
+        else:
+            method = '--no-ff'
+
+        rc, output = execute(
+            ['git', 'merge', target, method, '--no-commit'],
+            ignore_errors=True,
+            return_error_code=True)
+
+        if rc:
+            raise MergeError("Could not merge branch '%s' into '%s'.\n\n%s" %
+                             (target, destination, output))
+
+        self.create_commit(message, author, run_editor)
+
+    def push_upstream(self, remote_branch):
+        """Pushes the current branch to upstream."""
+        origin_url = self.get_origin()[1]
+        rc, output = execute(
+            ['git', 'pull', '--rebase', origin_url, remote_branch],
+            ignore_errors=True,
+            return_error_code=True)
+
+        if rc:
+            raise PushError('Could not pull changes from upstream.')
+
+        rc, output = execute(
+            ['git', 'push', origin_url, remote_branch],
+            ignore_errors=True,
+            return_error_code=True)
+
+        if rc:
+            raise PushError("Could not push branch '%s' to upstream" %
+                            remote_branch)
+
     def get_current_branch(self):
         """Returns the name of the current branch."""
         return execute([self.git, "rev-parse", "--abbrev-ref", "HEAD"],
                        ignore_errors=True).strip()
+
+    def _get_root_directory(self):
+        """Get the root directory of the repository as an absolute path."""
+        git_dir = execute([self.git, "rev-parse", "--git-dir"],
+                          ignore_errors=True).rstrip("\n")
+
+        if git_dir.startswith("fatal:") or not os.path.isdir(git_dir):
+            logging.error("Could not find git repository path.")
+            return None
+
+        return os.path.abspath(os.path.join(git_dir, ".."))
+
+    @property
+    def original_cwd(self):
+        """Get the original current working directory."""
+        if self._original_cwd is None:
+            # If this is None, then we haven't called get_repository_info and
+            # shouldn't have changed directories.
+            self._original_cwd = os.getcwd()
+
+        return self._original_cwd
