@@ -3,14 +3,15 @@ import os
 import re
 import uuid
 
-from urlparse import urlsplit, urlunparse
+from six.moves.urllib.parse import urlsplit, urlunparse
 
-from rbtools.clients import SCMClient, RepositoryInfo
+from rbtools.clients import PatchResult, SCMClient, RepositoryInfo
 from rbtools.clients.errors import (InvalidRevisionSpecError,
                                     TooManyRevisionsError)
 from rbtools.clients.svn import SVNClient
 from rbtools.utils.checks import check_install
 from rbtools.utils.filesystem import make_empty_files
+from rbtools.utils.console import edit_text
 from rbtools.utils.process import die, execute
 
 
@@ -24,6 +25,8 @@ class MercurialClient(SCMClient):
     PRE_CREATION = '/dev/null'
     PRE_CREATION_DATE = 'Thu Jan 01 00:00:00 1970 +0000'
 
+    supports_diff_exclude_patterns = True
+
     def __init__(self, **kwargs):
         super(MercurialClient, self).__init__(**kwargs)
 
@@ -230,6 +233,13 @@ class MercurialClient(SCMClient):
                 result['base'] = self._identify_revision(outgoing[0])
                 result['tip'] = self._identify_revision(outgoing[1])
                 result['commit_id'] = result['tip']
+                # Since the user asked us to operate on tip, warn them about a
+                # dirty working directory
+                if self.has_pending_changes():
+                    logging.warning('Your working directory is not clean. Any '
+                                    'changes which have not been committed '
+                                    'to a branch will not be included in your '
+                                    'review request.')
 
             if self.options.parent_branch:
                 result['parent_base'] = result['base']
@@ -262,7 +272,7 @@ class MercurialClient(SCMClient):
             # If there are missing changesets between base and the remote, we
             # need to generate a parent diff.
             outgoing = self._get_outgoing_changesets(self._get_remote_branch(),
-                                                 rev=result['base'])
+                                                     rev=result['base'])
 
             logging.debug('%d outgoing changesets between remote and base.',
                           len(outgoing))
@@ -270,10 +280,19 @@ class MercurialClient(SCMClient):
             if not outgoing:
                 return result
 
-            result['parent_base'] = self._execute(
+            parent_base = self._execute(
                 ['hg', 'parents', '--hidden', '-r', outgoing[0][1],
-                 '--template', '{node|short}']).split()[0]
-            logging.debug('Identified %s as parent base', result['parent_base'])
+                 '--template', '{node|short}']).split()
+
+            if len(parent_base) == 0:
+                raise Exception(
+                    'Could not find parent base revision. Ensure upstream '
+                    'repository is not empty.')
+
+            result['parent_base'] = parent_base[0]
+
+            logging.debug('Identified %s as parent base',
+                          result['parent_base'])
 
         return result
 
@@ -313,7 +332,6 @@ class MercurialClient(SCMClient):
         return RepositoryInfo(path=root, base_path=base_path,
                               supports_parent_diffs=True)
 
-
     def _load_hgrc(self):
         for line in execute(['hg', 'showconfig'], split_lines=True):
             line = line.split('=', 1)
@@ -344,7 +362,8 @@ class MercurialClient(SCMClient):
 
         return '\n\n'.join([desc.strip() for desc in descs])
 
-    def diff(self, revisions, files=[], extra_args=[]):
+    def diff(self, revisions, include_files=[], exclude_patterns=[],
+             extra_args=[]):
         """
         Performs a diff across all modified files in a Mercurial repository.
         """
@@ -355,26 +374,35 @@ class MercurialClient(SCMClient):
         if self._type == 'svn':
             diff_cmd.append('--svn')
 
-        diff_cmd += files
+        diff_cmd += include_files
+
+        for pattern in exclude_patterns:
+            diff_cmd.append('-X')
+            diff_cmd.append(pattern)
 
         diff = self._execute(
             diff_cmd + ['-r', revisions['base'], '-r', revisions['tip']],
-            env=self._hg_env)
+            env=self._hg_env, log_output_on_error=False, results_unicode=False)
 
-        if self._supports_empty_files():
+        supports_empty_files = self.supports_empty_files()
+
+        if supports_empty_files:
             diff = self._handle_empty_files(diff, revisions['base'],
-                                            revisions['tip'])
+                                            revisions['tip'],
+                                            exclude_files=exclude_patterns)
 
         if 'parent_base' in revisions:
             base_commit_id = revisions['parent_base']
             parent_diff = self._execute(
                 diff_cmd + ['-r', base_commit_id, '-r', revisions['base']],
-                env=self._hg_env)
+                env=self._hg_env, results_unicode=False)
 
-            if self._supports_empty_files():
-                parent_diff = self._handle_empty_files(parent_diff,
-                                                       base_commit_id,
-                                                       revisions['base'])
+            if supports_empty_files:
+                parent_diff = self._handle_empty_files(
+                    parent_diff,
+                    base_commit_id,
+                    revisions['base'],
+                    exclude_files=exclude_patterns)
         else:
             base_commit_id = revisions['base']
             parent_diff = None
@@ -386,7 +414,7 @@ class MercurialClient(SCMClient):
             'base_commit_id': base_commit_id,
         }
 
-    def _handle_empty_files(self, diff, base, tip):
+    def _handle_empty_files(self, diff, base, tip, exclude_files=[]):
         """Adds added and deleted 0-length files to the diff output.
 
         Since the diff output from hg diff does not give any information on
@@ -414,20 +442,22 @@ class MercurialClient(SCMClient):
                                       env=self._hg_env).strip().split('\t')
 
         for filename in added_empty_files:
-            diff += ('diff -r %s -r %s %s\n'
-                     '--- %s\t%s\n'
-                     '+++ b/%s\t%s\n'
-                     % (base, tip, filename,
-                        self.PRE_CREATION, self.PRE_CREATION_DATE,
-                        filename, tip_date))
+            if filename not in exclude_files:
+                diff += ('diff -r %s -r %s %s\n'
+                         '--- %s\t%s\n'
+                         '+++ b/%s\t%s\n'
+                         % (base, tip, filename,
+                            self.PRE_CREATION, self.PRE_CREATION_DATE,
+                            filename, tip_date))
 
         for filename in deleted_empty_files:
-            diff += ('diff -r %s -r %s %s\n'
-                     '--- a/%s\t%s\n'
-                     '+++ %s\t%s\n'
-                     % (base, tip, filename,
-                        filename, base_date,
-                        self.PRE_CREATION, self.PRE_CREATION_DATE))
+            if filename not in exclude_files:
+                diff += ('diff -r %s -r %s %s\n'
+                         '--- a/%s\t%s\n'
+                         '+++ %s\t%s\n'
+                         % (base, tip, filename,
+                            filename, base_date,
+                            self.PRE_CREATION, self.PRE_CREATION_DATE))
 
         return diff
 
@@ -463,15 +493,13 @@ class MercurialClient(SCMClient):
         If the remote branch is not defined, the parent branch of the
         repository is returned.
         """
-        try:
-            remote = self._remote_path[0]
-        except IndexError:
-            remote = None
+        remote = getattr(self.options, 'tracking', None)
 
-        tracking = getattr(self.options, 'tracking', None)
-
-        if not remote and tracking:
-            remote = tracking
+        if not remote:
+            try:
+                remote = self._remote_path[0]
+            except IndexError:
+                remote = None
 
         if not remote:
             die('Could not determine remote branch to use for diff creation. '
@@ -479,6 +507,24 @@ class MercurialClient(SCMClient):
 
         return remote
 
+    def create_commit(self, message, author, run_editor,
+                      files=[], all_files=False):
+        """Commits the given modified files.
+
+        This is expected to be called after applying a patch. This commits the
+        patch using information from the review request, opening the commit
+        message in $EDITOR to allow the user to update it.
+        """
+        if run_editor:
+            modified_message = edit_text(message)
+        else:
+            modified_message = message
+
+        hg_command = ['hg', 'commit', '-m', modified_message,
+                      '-u %s <%s>' % (author.fullname, author.email)]
+
+        execute(hg_command + files)
+
     def _get_current_branch(self):
         """Returns the current branch of this repository."""
         return execute(['hg', 'branch'], env=self._hg_env).strip()
@@ -608,20 +654,35 @@ class MercurialClient(SCMClient):
 
         return execute(cmd, *args, **kwargs)
 
-    def apply_patch(self, patch_file, base_path=None, base_dir=None, p=None):
+    def has_pending_changes(self):
+        """Checks if there are changes waiting to be committed.
+
+        Returns True if the working directory has been modified,
+        otherwise returns False.
+        """
+        status = execute(['hg', 'status', '--modified', '--added',
+                          '--removed', '--deleted'])
+        return status != ''
+
+    def apply_patch(self, patch_file, base_path=None, base_dir=None, p=None,
+                    revert=False):
         """Import the given patch.
 
         This will take the given patch file and apply it to the working
         directory.
         """
+        cmd = ['hg', 'patch', '--no-commit']
+
         if p:
-            cmd = ['hg', 'patch', '--no-commit', '-p', p, patch_file]
-        else:
-            cmd = ['hg', 'patch', '--no-commit', patch_file]
+            cmd += ['-p', p]
 
-        self._execute(cmd)
+        cmd.append(patch_file)
 
-    def _apply_patch_for_empty_files(self, patch, p_num):
+        rc, data = self._execute(cmd, with_errors=True, return_error_code=True)
+
+        return PatchResult(applied=(rc == 0), patch_output=data)
+
+    def apply_patch_for_empty_files(self, patch, p_num, revert=False):
         """Returns True if any empty files in the patch are applied.
 
         If there are no empty files in the patch or if an error occurs while
@@ -666,7 +727,7 @@ class MercurialClient(SCMClient):
 
         return patched_empty_files
 
-    def _supports_empty_files(self):
+    def supports_empty_files(self):
         """Checks if the RB server supports added/deleted empty files."""
         return (self.capabilities and
                 self.capabilities.has_capability('scmtools', 'mercurial',
