@@ -1,14 +1,21 @@
+from __future__ import print_function, unicode_literals
+
+import fnmatch
 import logging
 import marshal
 import os
 import re
+import six
 import socket
 import stat
+import string
 import subprocess
 
 from rbtools.clients import SCMClient, RepositoryInfo
-from rbtools.clients.errors import (EmptyChangeError,
+from rbtools.clients.errors import (AmendError,
+                                    EmptyChangeError,
                                     InvalidRevisionSpecError,
+                                    SCMError,
                                     TooManyRevisionsError)
 from rbtools.utils.checks import check_gnu_diff, check_install
 from rbtools.utils.filesystem import make_empty_files, make_tempfile
@@ -35,11 +42,15 @@ class P4Wrapper(object):
         lines = self.run_p4(['counters'], split_lines=True)
         return self._parse_keyval_lines(lines, self.COUNTERS_RE)
 
-    def change(self, changenum, password=None):
+    def change(self, changenum, marshalled=True, password=None):
         return self.run_p4(['change', '-o', str(changenum)],
                            password=password, ignore_errors=True,
                            none_on_ignored_error=True,
-                           marshalled=True)
+                           marshalled=marshalled)
+
+    def modify_change(self, new_change_spec):
+        """new_change_spec must contain the changelist number."""
+        return self.run_p4(['change', '-i'], input_string=new_change_spec)
 
     def files(self, path):
         return self.run_p4(['files', path], marshalled=True)
@@ -92,7 +103,13 @@ class P4Wrapper(object):
         return self.run_p4(['where', depot_path], marshalled=True)
 
     def run_p4(self, p4_args, marshalled=False, password=None,
-               ignore_errors=False, *args, **kwargs):
+               ignore_errors=False, input_string=None, *args, **kwargs):
+        """Invoke p4.
+
+        In the current implementation, the arguments 'marshalled' and
+        'input_string' cannot be used together, i.e. this command doesn't
+        allow inputting and outputting at the same time.
+        """
         cmd = ['p4']
 
         if marshalled:
@@ -132,10 +149,21 @@ class P4Wrapper(object):
             if not ignore_errors and (rc or has_error):
                 for record in result:
                     if 'data' in record:
-                        print record['data']
-                die('Failed to execute command: %s\n' % (cmd,))
+                        print(record['data'])
+
+                raise SCMError('Failed to execute command: %s\n' % cmd)
 
             return result
+
+        elif input_string is not None:
+            p = subprocess.Popen(cmd, stdin=subprocess.PIPE)
+            p.communicate(input_string)  # Send input, wait, set returncode
+
+            if not ignore_errors and p.returncode:
+                raise SCMError('Failed to execute command: %s\n' % cmd)
+
+            return None
+
         else:
             result = execute(cmd, ignore_errors=ignore_errors, *args, **kwargs)
 
@@ -162,15 +190,24 @@ class PerforceClient(SCMClient):
     """
     name = 'Perforce'
 
+    can_amend_commit = True
+    supports_diff_exclude_patterns = True
     supports_diff_extra_args = True
+    supports_patch_revert = True
 
-    DATE_RE = re.compile(r'(\w+)\s+(\w+)\s+(\d+)\s+(\d\d:\d\d:\d\d)\s+'
-                         '(\d\d\d\d)')
+    DATE_RE = re.compile(br'(\w+)\s+(\w+)\s+(\d+)\s+(\d\d:\d\d:\d\d)\s+'
+                         br'(\d\d\d\d)')
     ENCODED_COUNTER_URL_RE = re.compile('reviewboard.url\.(\S+)')
 
     REVISION_CURRENT_SYNC = '--rbtools-current-sync'
+    REVISION_DEFAULT_CLN = 'default'
     REVISION_PENDING_CLN_PREFIX = '--rbtools-pending-cln:'
 
+    ADDED_FILES_RE = re.compile(r'^==== //depot/(\S+)#\d+ ==A== \S+ ====$',
+                                re.M)
+    DELETED_FILES_RE = re.compile(r'^==== //depot/(\S+)#\d+ ==D== \S+ ====$',
+                                  re.M)
+
     def __init__(self, p4_class=P4Wrapper, **kwargs):
         super(PerforceClient, self).__init__(**kwargs)
         self.p4 = p4_class(self.options)
@@ -231,7 +268,7 @@ class PerforceClient(SCMClient):
             if info[1]:
                 servers += info[1]
 
-            repository_path = ["%s:%s" % (server, port)
+            repository_path = ['%s:%s' % (server, port)
                                for server in servers]
 
             # If there's only one repository path found, then we don't
@@ -293,7 +330,8 @@ class PerforceClient(SCMClient):
         if n_revs == 0:
             return {
                 'base': self.REVISION_CURRENT_SYNC,
-                'tip': self.REVISION_PENDING_CLN_PREFIX + 'default',
+                'tip': (self.REVISION_PENDING_CLN_PREFIX +
+                        self.REVISION_DEFAULT_CLN)
             }
         elif n_revs == 1:
             # A single specified CLN can be any of submitted, pending, or
@@ -371,7 +409,7 @@ class PerforceClient(SCMClient):
             raise TooManyRevisionsError
 
     def _get_changelist_status(self, changelist):
-        if changelist == 'default':
+        if changelist == self.REVISION_DEFAULT_CLN:
             return 'pending'
         else:
             change = self.p4.change(changelist)
@@ -412,7 +450,7 @@ class PerforceClient(SCMClient):
 
         # Next try for a counter of the form:
         # reviewboard_url.http:||reviewboard.example.com
-        for key, value in counters.iteritems():
+        for key, value in six.iteritems(counters):
             m = self.ENCODED_COUNTER_URL_RE.match(key)
 
             if m:
@@ -420,23 +458,26 @@ class PerforceClient(SCMClient):
 
         return None
 
-    def diff(self, revisions, files=[], extra_args=[]):
+    def diff(self, revisions, include_files=[], exclude_patterns=[],
+             extra_args=[]):
         """
         Goes through the hard work of generating a diff on Perforce in order
         to take into account adds/deletes and to provide the necessary
         revision information.
         """
+        exclude_patterns = self.normalize_exclude_patterns(exclude_patterns)
+
         if not revisions:
             # The "path posting" is still interesting enough to keep around. If
             # the given arguments don't parse as valid changelists, fall back
             # on that behavior.
-            return self._path_diff(extra_args)
+            return self._path_diff(extra_args, exclude_patterns)
 
         # Support both //depot/... paths and local filenames. For the moment,
         # this does *not* support any of perforce's traversal literals like ...
         depot_include_files = []
         local_include_files = []
-        for filename in files:
+        for filename in include_files:
             if filename.startswith('//'):
                 depot_include_files.append(filename)
             else:
@@ -457,10 +498,11 @@ class PerforceClient(SCMClient):
                          'to %s',
                          base, tip)
             return self._compute_range_changes(
-                base, tip, depot_include_files, local_include_files)
+                base, tip, depot_include_files, local_include_files,
+                exclude_patterns)
 
         # Strip off the prefix
-        tip = tip[len(self.REVISION_PENDING_CLN_PREFIX):]
+        tip = tip.split(':', 1)[1]
 
         # Try to get the files out of the working directory first. If that
         # doesn't work, look at shelved files.
@@ -484,6 +526,7 @@ class PerforceClient(SCMClient):
             'integrate': 'M',
             'add': 'A',
             'branch': 'A',
+            'import': 'A',
             'delete': 'D',
         }
 
@@ -518,7 +561,9 @@ class PerforceClient(SCMClient):
             if ((depot_include_files and
                  depot_file not in depot_include_files) or
                 (local_include_files and
-                 local_file not in local_include_files)):
+                 local_file not in local_include_files) or
+                self._should_exclude_file(local_file, depot_file,
+                                          exclude_patterns)):
                 continue
 
             old_file = ''
@@ -529,14 +574,15 @@ class PerforceClient(SCMClient):
             try:
                 changetype_short = action_mapping[action]
             except KeyError:
-                die('Unsupported action type "%s" for %s' % (action, depot_file))
+                die('Unsupported action type "%s" for %s'
+                    % (action, depot_file))
 
             if changetype_short == 'M':
                 try:
                     old_file, new_file = self._extract_edit_files(
                         depot_file, local_file, base_revision, tip,
                         cl_is_shelved, False)
-                except ValueError, e:
+                except ValueError as e:
                     logging.warning('Skipping file %s: %s', depot_file, e)
                     continue
             elif changetype_short == 'A':
@@ -556,7 +602,7 @@ class PerforceClient(SCMClient):
                     old_file, new_file = self._extract_add_files(
                         depot_file, local_file, tip, cl_is_shelved,
                         cl_is_pending)
-                except ValueError, e:
+                except ValueError as e:
                     logging.warning('Skipping file %s: %s', depot_file, e)
                     continue
 
@@ -567,7 +613,7 @@ class PerforceClient(SCMClient):
                 try:
                     old_file, new_file = self._extract_delete_files(
                         depot_file, base_revision)
-                except ValueError, e:
+                except ValueError as e:
                     logging.warning('Skipping file %s#%s: %s', depot_file, e)
                     continue
             elif changetype_short == 'MV-a':
@@ -580,7 +626,7 @@ class PerforceClient(SCMClient):
                     old_file, new_file, new_depot_file = \
                         self._extract_move_files(
                             depot_file, tip, base_revision, cl_is_shelved)
-                except ValueError, e:
+                except ValueError as e:
                     logging.warning('Skipping file %s: %s', depot_file, e)
                     continue
 
@@ -594,18 +640,18 @@ class PerforceClient(SCMClient):
         # description server-side. Ideally we'd change this to remove the
         # server-side implementation and just implement --guess-summary and
         # --guess-description, but that would create a lot of unhappy users.
-        if cl_is_pending and tip != 'default':
+        if cl_is_pending and tip != self.REVISION_DEFAULT_CLN:
             changenum = str(tip)
         else:
             changenum = None
 
         return {
-            'diff': ''.join(diff_lines),
+            'diff': b''.join(diff_lines),
             'changenum': changenum,
         }
 
     def _compute_range_changes(self, base, tip, depot_include_files,
-                               local_include_files):
+                               local_include_files, exclude_patterns):
         """Compute the changes across files given a revision range.
 
         This will look at the history of all changes within the given range and
@@ -680,7 +726,7 @@ class PerforceClient(SCMClient):
         files = []
         for cln in sorted(changesets.keys()):
             changeset = changesets[cln]
-            for depot_file, change in changeset.iteritems():
+            for depot_file, change in six.iteritems(changeset):
                 action = change['action']
 
                 # Moves will be handled in the 'move/delete' entry
@@ -711,6 +757,7 @@ class PerforceClient(SCMClient):
         # Now generate the diff
         supports_moves = self._supports_moves()
         diff_lines = []
+
         for f in files:
             action = f['action']
             depot_file = f['depotFile']
@@ -722,14 +769,16 @@ class PerforceClient(SCMClient):
             if ((depot_include_files and
                  depot_file not in depot_include_files) or
                 (local_include_files and
-                 local_file not in local_include_files)):
+                 local_file not in local_include_files) or
+                self._should_exclude_file(local_file, depot_file,
+                                          exclude_patterns)):
                 continue
 
             if action == 'add':
                 try:
                     old_file, new_file = self._extract_add_files(
                         depot_file, local_file, rev, False, False)
-                except ValueError, e:
+                except ValueError as e:
                     logging.warning('Skipping file %s: %s', depot_file, e)
                     continue
 
@@ -779,8 +828,8 @@ class PerforceClient(SCMClient):
                         old_file_a, new_file_a, depot_file, 0, '', 'A',
                         ignore_unmodified=True)
                     diff_lines += self._do_diff(
-                        old_file_b, new_file_b, initial_depot_file, initial_rev,
-                        depot_file, 'D', ignore_unmodified=True)
+                        old_file_b, new_file_b, initial_depot_file,
+                        initial_rev, depot_file, 'D', ignore_unmodified=True)
             elif action == 'skip':
                 continue
             else:
@@ -790,7 +839,7 @@ class PerforceClient(SCMClient):
                 assert False
 
         return {
-            'diff': ''.join(diff_lines)
+            'diff': b''.join(diff_lines)
         }
 
     def _accumulate_range_change(self, file_entry, change):
@@ -803,9 +852,9 @@ class PerforceClient(SCMClient):
             new_action = current_action
             file_entry['depotFile'] = file_entry['initialDepotFile']
 
-            # If the first action was an edit, then the initial revision
-            # (that we'll use to generate the diff) is n-1
-            if current_action == 'edit':
+            # If the first action was an edit or a delete, then the initial
+            # revision (that we'll use to generate the diff) is n-1
+            if current_action in ('edit', 'delete'):
                 file_entry['initialRev'] -= 1
         elif current_action == 'add':
             # If we're adding a file that existed in the base changeset, it
@@ -942,7 +991,7 @@ class PerforceClient(SCMClient):
 
         return old_filename, new_filename, new_depot_file
 
-    def _path_diff(self, args):
+    def _path_diff(self, args, exclude_patterns):
         """
         Process a path-style diff. This allows people to post individual files
         in various ways.
@@ -1021,7 +1070,8 @@ class PerforceClient(SCMClient):
             old_file = new_file = empty_filename
             changetype_short = None
 
-            for depot_path, (first_record, second_record) in files.items():
+            for depot_path, (first_record, second_record) in \
+                    six.iteritems(files):
                 old_file = new_file = empty_filename
                 if first_record is None:
                     new_path = '%s#%s' % (depot_path, second_record['rev'])
@@ -1050,6 +1100,11 @@ class PerforceClient(SCMClient):
                     changetype_short = 'M'
                     base_revision = int(first_record['rev'])
 
+                local_path = self._depot_to_local(depot_path)
+                if self._should_exclude_file(local_path, depot_path,
+                                             exclude_patterns):
+                    continue
+
                 # TODO: We're passing new_depot_file='' here just to make
                 # things work like they did before the moved file change was
                 # added (58ccae27). This section of code needs to be updated
@@ -1064,7 +1119,7 @@ class PerforceClient(SCMClient):
         os.unlink(tmp_diff_to_filename)
 
         return {
-            'diff': ''.join(diff_lines),
+            'diff': b''.join(diff_lines),
         }
 
     def _do_diff(self, old_file, new_file, depot_file, base_revision,
@@ -1085,16 +1140,17 @@ class PerforceClient(SCMClient):
         Returns a list of strings of diff lines.
         """
         if hasattr(os, 'uname') and os.uname()[0] == 'SunOS':
-            diff_cmd = ["gdiff", "-urNp", old_file, new_file]
+            diff_cmd = ['gdiff', '-urNp', old_file, new_file]
         else:
-            diff_cmd = ["diff", "-urNp", old_file, new_file]
+            diff_cmd = ['diff', '-urNp', old_file, new_file]
 
         # Diff returns "1" if differences were found.
         dl = execute(diff_cmd, extra_ignore_errors=(1, 2),
-                     translate_newlines=False)
+                     log_output_on_error=False, translate_newlines=False,
+                     results_unicode=False)
 
         # If the input file has ^M characters at end of line, lets ignore them.
-        dl = dl.replace('\r\r\n', '\r\n')
+        dl = dl.replace(b'\r\r\n', b'\r\n')
         dl = dl.splitlines(True)
 
         cwd = os.getcwd()
@@ -1120,69 +1176,74 @@ class PerforceClient(SCMClient):
         # and the code below expects the output to start with
         #     "Binary files "
         if (len(dl) == 1 and
-            dl[0].startswith('Files %s and %s differ' %
-                            (old_file, new_file))):
-            dl = ['Binary files %s and %s differ\n' % (old_file, new_file)]
+            dl[0].startswith(b'Files %s and %s differ' %
+                             (old_file, new_file))):
+            dl = [b'Binary files %s and %s differ\n' % (old_file, new_file)]
 
-        if dl == [] or dl[0].startswith("Binary files "):
-            is_empty_and_changed = (self._supports_empty_files() and
+        if dl == [] or dl[0].startswith(b'Binary files '):
+            is_empty_and_changed = (self.supports_empty_files() and
                                     changetype_short in ('A', 'D'))
 
             if dl == [] and (is_move or is_empty_and_changed):
-                dl.insert(0, "==== %s#%s ==%s== %s ====\n" %
-                          (depot_file, base_revision, changetype_short,
-                           new_local_path))
-                dl.append('\n')
+                line = ('==== %s#%s ==%s== %s ====\n'
+                        % (depot_file, base_revision, changetype_short,
+                           new_local_path)).encode('utf-8')
+                dl.insert(0, line)
+                dl.append(b'\n')
             else:
                 if ignore_unmodified:
                     return []
                 else:
-                    print "Warning: %s in your changeset is unmodified" % \
-                          local_path
+                    print('Warning: %s in your changeset is unmodified' %
+                          local_path)
         elif len(dl) > 1:
-            m = re.search(r'(\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d)', dl[1])
+            m = re.search(br'(\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d)', dl[1])
             if m:
-                timestamp = m.group(1)
+                timestamp = m.group(1).decode('utf-8')
             else:
                 # Thu Sep  3 11:24:48 2007
                 m = self.DATE_RE.search(dl[1])
                 if not m:
-                    die("Unable to parse diff header: %s" % dl[1])
+                    die('Unable to parse diff header: %s' % dl[1])
 
                 month_map = {
-                    "Jan": "01",
-                    "Feb": "02",
-                    "Mar": "03",
-                    "Apr": "04",
-                    "May": "05",
-                    "Jun": "06",
-                    "Jul": "07",
-                    "Aug": "08",
-                    "Sep": "09",
-                    "Oct": "10",
-                    "Nov": "11",
-                    "Dec": "12",
+                    b'Jan': b'01',
+                    b'Feb': b'02',
+                    b'Mar': b'03',
+                    b'Apr': b'04',
+                    b'May': b'05',
+                    b'Jun': b'06',
+                    b'Jul': b'07',
+                    b'Aug': b'08',
+                    b'Sep': b'09',
+                    b'Oct': b'10',
+                    b'Nov': b'11',
+                    b'Dec': b'12',
                 }
                 month = month_map[m.group(2)]
                 day = m.group(3)
                 timestamp = m.group(4)
                 year = m.group(5)
 
-                timestamp = "%s-%s-%s %s" % (year, month, day, timestamp)
+                timestamp = '%s-%s-%s %s' % (year, month, day, timestamp)
 
-            dl[0] = "--- %s\t%s#%s\n" % (local_path, depot_file, base_revision)
-            dl[1] = "+++ %s\t%s\n" % (new_local_path, timestamp)
+            dl[0] = ('--- %s\t%s#%s\n'
+                     % (local_path, depot_file, base_revision)).encode('utf-8')
+            dl[1] = ('+++ %s\t%s\n'
+                     % (new_local_path, timestamp)).encode('utf-8')
 
             if is_move:
-                dl.insert(0, 'Moved to: %s\n' % new_depot_file)
-                dl.insert(0, 'Moved from: %s\n' % depot_file)
+                dl.insert(0,
+                          ('Moved to: %s\n' % new_depot_file).encode('utf-8'))
+                dl.insert(0,
+                          ('Moved from: %s\n' % depot_file).encode('utf-8'))
 
             # Not everybody has files that end in a newline (ugh). This ensures
             # that the resulting diff file isn't broken.
-            if dl[-1][-1] != '\n':
-                dl.append('\n')
+            if not dl[-1].endswith(b'\n'):
+                dl.append(b'\n')
         else:
-            die("ERROR, no valid diffs: %s" % dl[0])
+            die('ERROR, no valid diffs: %s' % dl[0].decode('utf-8'))
 
         return dl
 
@@ -1207,7 +1268,7 @@ class PerforceClient(SCMClient):
         #   when rbt uses their credentials to publish its contents.
 
         if os.path.islink(tmpfile):
-            raise ValueError("'%s' is a symlink" % depot_path)
+            raise ValueError('"%s" is a symlink' % depot_path)
         else:
             os.chmod(tmpfile, stat.S_IREAD | stat.S_IWRITE)
 
@@ -1225,17 +1286,46 @@ class PerforceClient(SCMClient):
             # XXX: This breaks on filenames with spaces.
             return where_output[-1]['data'].split(' ')[2].strip()
 
-    def _apply_patch_for_empty_files(self, patch, p_num):
+    def get_raw_commit_message(self, revisions):
+        """Extract the commit message based on the provided revision range.
+
+        Since local changelists in perforce are not ordered with respect to
+        one another, this implementation looks at only the tip revision.
+        """
+        changelist = revisions['tip']
+
+        # The parsed revision spec may include a prefix indicating that it is
+        # pending. This prefix, which is delimited by a colon, must be
+        # stripped in order to run p4 change on the actual changelist number.
+        changelist = changelist.split(':', 1)[1]
+
+        if changelist == self.REVISION_DEFAULT_CLN:
+            # The default changelist has no description and couldn't be
+            # accessed from p4 change anyway
+            return ''
+
+        logging.debug('Fetching description for changelist %s', changelist)
+        change = self.p4.change(changelist)
+
+        if len(change) == 1 and 'Description' in change[0]:
+            return change[0]['Description']
+        else:
+            return ''
+
+    def apply_patch_for_empty_files(self, patch, p_num, revert=False):
         """Returns True if any empty files in the patch are applied.
 
         If there are no empty files in the patch or if an error occurs while
         applying the patch, we return False.
         """
         patched_empty_files = False
-        added_files = re.findall(r'^==== //depot/(\S+)#\d+ ==A== \S+ ====$',
-                                 patch, re.M)
-        deleted_files = re.findall(r'^==== //depot/(\S+)#\d+ ==D== \S+ ====$',
-                                   patch, re.M)
+
+        if revert:
+            added_files = self.DELETED_FILES_RE.findall(patch)
+            deleted_files = self.ADDED_FILES_RE.findall(patch)
+        else:
+            added_files = self.ADDED_FILES_RE.findall(patch)
+            deleted_files = self.DELETED_FILES_RE.findall(patch)
 
         # Prepend the root of the Perforce client to each file name.
         p4_info = self.p4.info()
@@ -1276,3 +1366,122 @@ class PerforceClient(SCMClient):
         return (self.capabilities and
                 self.capabilities.has_capability('scmtools', 'perforce',
                                                  'empty_files'))
+
+    def _should_exclude_file(self, local_file, depot_file, exclude_patterns):
+        """Determine if a file should be excluded from a diff.
+
+        Check if the file identified by (local_file, depot_file) should be
+        excluded from the diff. If a pattern beings with '//', then it will be
+        matched against the depot_file. Otherwise, it will be matched against
+        the local file.
+
+        This function expects `exclude_patterns` to be normalized.
+        """
+        for pattern in exclude_patterns:
+            if pattern.startswith('//'):
+                if fnmatch.fnmatch(depot_file, pattern):
+                    return True
+            elif fnmatch.fnmatch(local_file, pattern):
+                    return True
+
+        return False
+
+    def normalize_exclude_patterns(self, patterns):
+        """Normalize the set of patterns so all non-depot paths are absolute.
+
+        A path with a leading // is interpreted as a depot pattern and remains
+        unchanged. A path with a leading path separator is interpreted as being
+        relative to the Perforce client root. All other paths are interpreted
+        as being relative to the current working directory. Non-depot paths are
+        transformed into absolute paths.
+        """
+        cwd = os.getcwd()
+        base_dir = self.p4.info().get('Client root')
+
+        def normalize(p):
+            if p.startswith('//'):
+                # Absolute depot patterns remain unchanged.
+                return p
+            elif pattern.startswith(os.path.sep):
+                # Patterns beginning with the operating system's path separator
+                # are relative to the repository root.
+                assert base_dir is not None
+                p = os.path.join(base_dir, p[1:])
+            else:
+                # All other patterns are considered to be relative to the
+                # current working directory.
+                p = os.path.join(cwd, p)
+
+            return os.path.normpath(p)
+
+        return [normalize(pattern) for pattern in patterns]
+
+    def _replace_description_in_changelist_spec(self, old_spec,
+                                                new_description):
+        """Replace the description in the given changelist spec.
+
+        old_spec is a formatted p4 changelist spec string (the raw output from
+        p4 change). This method replaces the existing description with
+        new_description, and returns the new changelist spec.
+        """
+        new_spec = ''
+        whitespace = tuple(string.whitespace)
+        description_key = 'Description:'
+        skipping_old_description = False
+
+        for line in old_spec.splitlines(True):
+            if not skipping_old_description:
+                if not line.startswith(description_key):
+                    new_spec += line
+
+                else:
+                    # Insert the new description. Don't include the first line
+                    # of the old one if it happens to be on the same line as
+                    # the key.
+                    skipping_old_description = True
+                    new_spec += description_key
+
+                    for desc_line in new_description.splitlines():
+                        new_spec += '\t%s\n' % desc_line
+
+            else:
+                # Ignore the description from the original file (all lines
+                # that start with whitespace until the next key is
+                # encountered).
+                if line.startswith(whitespace):
+                    continue
+                else:
+                    skipping_old_description = False
+                    new_spec += '\n%s' % line
+
+        return new_spec
+
+    def amend_commit_description(self, message, revisions):
+        """Update a commit message to the given string.
+
+        Since local changelists on perforce have no ordering with respect to
+        each other, the revisions argument is mandatory.
+        """
+        # Get the changelist number from the tip revision, removing the prefix
+        # if necessary. Don't allow amending submitted or default changelists.
+        changelist_id = revisions['tip']
+        logging.debug('Preparing to amend change %s' % changelist_id)
+
+        if not changelist_id.startswith(self.REVISION_PENDING_CLN_PREFIX):
+            raise AmendError('Cannot modify submitted changelist %s'
+                             % changelist_id)
+
+        changelist_num = changelist_id.split(':', 1)[1]
+
+        if changelist_num == self.REVISION_DEFAULT_CLN:
+            raise AmendError('Cannot modify the default changelist')
+        elif not changelist_num.isdigit():
+            raise AmendError('%s is an invalid changelist ID' % changelist_num)
+
+        # Get the current changelist description and insert the new message.
+        # Since p4 change -i doesn't take in marshalled objects, we get the
+        # description as raw text and manually edit it.
+        change = self.p4.change(changelist_num, marshalled=False)
+        new_change = self._replace_description_in_changelist_spec(
+            change, message)
+        self.p4.modify_change(new_change)
