@@ -1,3 +1,5 @@
+from __future__ import print_function, unicode_literals
+
 import argparse
 import getpass
 import inspect
@@ -5,7 +7,9 @@ import logging
 import platform
 import os
 import sys
-from urlparse import urlparse
+
+from six.moves import input
+from six.moves.urllib.parse import urlparse
 
 from rbtools import get_version_string
 from rbtools.api.capabilities import Capabilities
@@ -18,12 +22,12 @@ from rbtools.utils.filesystem import (cleanup_tempfiles, get_home_path,
 from rbtools.utils.process import die
 
 
-RB_MAIN = "rbt"
+RB_MAIN = 'rbt'
 
 
 class CommandExit(Exception):
     def __init__(self, exit_code=0):
-        super(CommandExit, self).__init__("Exit with code %s" % exit_code)
+        super(CommandExit, self).__init__('Exit with code %s' % exit_code)
         self.exit_code = exit_code
 
 
@@ -35,6 +39,24 @@ class ParseError(CommandError):
     pass
 
 
+class SmartHelpFormatter(argparse.HelpFormatter):
+    """Smartly formats help text, preserving paragraphs."""
+    def _split_lines(self, text, width):
+        # NOTE: This function depends on overriding _split_lines's behavior.
+        #       It is clearly documented that this function should not be
+        #       considered public API. However, given that the width we need
+        #       is calculated by HelpFormatter, and HelpFormatter has no
+        #       blessed public API, we have no other choice but to override
+        #       it here.
+        lines = []
+
+        for line in text.splitlines():
+            lines += super(SmartHelpFormatter, self)._split_lines(line, width)
+            lines.append('')
+
+        return lines[:-1]
+
+
 class Option(object):
     """Represents an option for a command.
 
@@ -62,13 +84,24 @@ class Option(object):
         Once we've determined the right set of flags, the option will be
         added to the parser.
         """
-        if 'config_key' in self.attrs:
-            if self.attrs['config_key'] in config:
-                self.attrs['default'] = config[self.attrs['config_key']]
+        attrs = self.attrs.copy()
 
-            del self.attrs['config_key']
+        if 'config_key' in attrs:
+            config_key = attrs.pop('config_key')
 
-        parent.add_argument(*self.opts, **self.attrs)
+            if config_key in config:
+                attrs['default'] = config[config_key]
+
+        if 'deprecated_in' in attrs:
+            attrs['help'] += '\n[Deprecated since %s]' % attrs['deprecated_in']
+
+        # These are used for other purposes, and are not supported by
+        # argparse.
+        for attr in ('added_in', 'deprecated_in', 'extended_help',
+                     'versions_changed'):
+            attrs.pop(attr, None)
+
+        parent.add_argument(*self.opts, **attrs)
 
 
 class OptionGroup(object):
@@ -127,18 +160,20 @@ class Command(object):
     ``option_list`` is a list of command line options for the command.
     Each list entry should be an Option or OptionGroup instance.
     """
-    name = ""
-    author = ""
-    description = ""
-    args = ""
+    name = ''
+    author = ''
+    description = ''
+    args = ''
     option_list = []
     _global_options = [
-        Option("-d", "--debug",
-               action="store_true",
-               dest="debug",
-               config_key="DEBUG",
+        Option('-d', '--debug',
+               action='store_true',
+               dest='debug',
+               config_key='DEBUG',
                default=False,
-               help="display debug output"),
+               help='Displays debug output.',
+               extended_help='This information can be valuable when debugging '
+                             'problems running the command.'),
     ]
 
     server_options = OptionGroup(
@@ -148,17 +183,10 @@ class Command(object):
         option_list=[
             Option('--server',
                    dest='server',
-                   metavar='SERVER',
+                   metavar='URL',
                    config_key='REVIEWBOARD_URL',
                    default=None,
                    help='Specifies the Review Board server to use.'),
-            Option('--disable-proxy',
-                   action='store_false',
-                   dest='enable_proxy',
-                   config_key='ENABLE_PROXY',
-                   default=True,
-                   help='Prevents requests from going through a proxy '
-                        'server.'),
             Option('--username',
                    dest='username',
                    metavar='USERNAME',
@@ -173,6 +201,61 @@ class Command(object):
                    default=None,
                    help='The password to be supplied to the Review Board '
                         'server.'),
+            Option('--api-token',
+                   dest='api_token',
+                   metavar='TOKEN',
+                   config_key='API_TOKEN',
+                   default=None,
+                   help='The API token to use for authentication, instead of '
+                        'using a username and password.',
+                   added_in='0.7'),
+            Option('--disable-proxy',
+                   action='store_false',
+                   dest='enable_proxy',
+                   config_key='ENABLE_PROXY',
+                   default=True,
+                   help='Prevents requests from going through a proxy '
+                        'server.'),
+            Option('--disable-ssl-verification',
+                   action='store_true',
+                   dest='disable_ssl_verification',
+                   config_key='DISABLE_SSL_VERIFICATION',
+                   default=False,
+                   help='Disable SSL certificate verification. This is useful '
+                        'with servers that have self-signed certificates.',
+                   added_in='0.7.3'),
+            Option('--disable-cookie-storage',
+                   config_key='SAVE_COOKIES',
+                   dest='save_cookies',
+                   action='store_false',
+                   default=True,
+                   help='Use an in-memory cookie store instead of writing '
+                        'them to a file. No credentials will be saved or '
+                        'loaded.',
+                   added_in='0.7.3'),
+            Option('--disable-cache',
+                   dest='disable_cache',
+                   config_key='DISABLE_CACHE',
+                   action='store_true',
+                   default=False,
+                   help='Disable the HTTP cache completely. This will '
+                        'result in slower requests.',
+                   added_in='0.7.3'),
+            Option('--disable-cache-storage',
+                   dest='in_memory_cache',
+                   config_key='IN_MEMORY_CACHE',
+                   action='store_true',
+                   default=False,
+                   help='Disable storing the API cache on the filesystem, '
+                        'instead keeping it in memory temporarily.',
+                   added_in='0.7.3'),
+            Option('--cache-location',
+                   dest='cache_location',
+                   metavar='FILE',
+                   config_key='CACHE_LOCATION',
+                   default=None,
+                   help='The file to use for the API cache database.',
+                   added_in='0.7.3'),
         ]
     )
 
@@ -181,24 +264,39 @@ class Command(object):
         option_list=[
             Option('--repository',
                    dest='repository_name',
+                   metavar='NAME',
                    config_key='REPOSITORY',
                    default=None,
                    help='The name of the repository configured on '
                         'Review Board that matches the local repository.'),
             Option('--repository-url',
                    dest='repository_url',
+                   metavar='URL',
                    config_key='REPOSITORY_URL',
                    default=None,
-                   help='The URL for a repository, used for creating '
-                        'a diff outside of a working copy (currently only '
-                        'supported by Subversion with specific revisions '
-                        'or --diff-filename and ClearCase with relative '
-                        'paths outside the view). For git, this specifies '
-                        'the origin url of the current repository, '
-                        'overriding the origin URL supplied by the git '
-                        'client.'),
+                   help='The URL for a repository.'
+                        '\n'
+                        'When generating diffs, this can be used for '
+                        'creating a diff outside of a working copy '
+                        '(currently only supported by Subversion with '
+                        'specific revisions or --diff-filename, and by '
+                        'ClearCase with relative paths outside the view).'
+                        '\n'
+                        'For Git, this specifies the origin URL of the '
+                        'current repository, overriding the origin URL '
+                        'supplied by the client.',
+                   versions_changed={
+                       '0.6': 'Prior versions used the `REPOSITORY` setting '
+                              'in .reviewboardrc, and allowed a '
+                              'repository name to be passed to '
+                              '--repository-url. This is no '
+                              'longer supported in 0.6 and higher. You '
+                              'may need to update your configuration and '
+                              'scripts appropriately.',
+                   }),
             Option('--repository-type',
                    dest='repository_type',
+                   metavar='TYPE',
                    config_key='REPOSITORY_TYPE',
                    default=None,
                    help='The type of repository in the current directory. '
@@ -218,15 +316,58 @@ class Command(object):
         option_list=[
             Option('--revision-range',
                    dest='revision_range',
+                   metavar='REV1:REV2',
                    default=None,
-                   help='Generates a diff for the given revision range. '
-                        '[DEPRECATED]'),
+                   help='Generates a diff for the given revision range.',
+                   deprecated_in='0.6'),
             Option('-I', '--include',
+                   metavar='FILENAME',
                    dest='include_files',
                    action='append',
-                   help='Includes only the given file in the diff. '
+                   help='Includes only the specified file in the diff. '
                         'This can be used multiple times to specify '
-                        'multiple files.'),
+                        'multiple files.'
+                        '\n'
+                        'Supported by: Bazaar, CVS, Git, Mercurial, '
+                        'Perforce, and Subversion.',
+                   added_in='0.6'),
+            Option('-X', '--exclude',
+                   metavar='PATTERN',
+                   dest='exclude_patterns',
+                   action='append',
+                   config_key='EXCLUDE_PATTERNS',
+                   help='Excludes all files that match the given pattern '
+                        'from the diff. This can be used multiple times to '
+                        'specify multiple patterns. UNIX glob syntax is used '
+                        'for pattern matching.'
+                        '\n'
+                        'Supported by: Bazaar, CVS, Git, Mercurial, '
+                        'Perforce, and Subversion.',
+                   extended_help=(
+                       'Patterns that begin with a path separator (/ on Mac '
+                       'OS and Linux, \\ on Windows) will be treated as being '
+                       'relative to the root of the repository. All other '
+                       'patterns are treated as being relative to the current '
+                       'working directory.'
+                       '\n'
+                       'For example, to exclude all ".txt" files from the '
+                       'resulting diff, you would use "-X /\'*.txt\'".'
+                       '\n'
+                       'When working with Mercurial, the patterns are '
+                       'provided directly to "hg" and are not limited to '
+                       'globs. For more information on advanced pattern '
+                       'syntax in Mercurial, run "hg help patterns"'
+                       '\n'
+                       'When working with CVS all diffs are generated '
+                       'relative to the current working directory so '
+                       'patterns beginning with a path separator are treated '
+                       'as relative to the current working directory.'
+                       '\n'
+                       'When working with Perforce, an exclude pattern '
+                       'beginning with `//` will be matched against depot '
+                       'paths; all other patterns will be matched against '
+                       'local paths.'),
+                   added_in='0.7'),
             Option('--parent',
                    dest='parent_branch',
                    metavar='BRANCH',
@@ -246,10 +387,12 @@ class Command(object):
                    config_key='TRACKING_BRANCH',
                    default=None,
                    help='The remote tracking branch from which your local '
-                        'branch is derived (Git/Mercurial only). Defaults '
-                        'are "origin/master" for Git and one of '
-                        '"reviewboard", "origin", "parent", or "default" for '
-                        'Mercurial.'),
+                        'branch is derived (Git/Mercurial only).'
+                        '\n'
+                        'For Git, the default is `origin/master`.'
+                        '\n'
+                        'For Mercurial, the default is one of: '
+                        '`reviewboard`, `origin`, `parent`, or `default`.'),
         ]
     )
 
@@ -294,17 +437,55 @@ class Command(object):
                    help='The path within the repository where the diff '
                         'was generated. This overrides the detected path. '
                         'Often used when passing --diff-filename.'),
+            Option('--svn-username',
+                   dest='svn_username',
+                   default=None,
+                   metavar='USERNAME',
+                   help='The username for the SVN repository.'),
+            Option('--svn-password',
+                   dest='svn_password',
+                   default=None,
+                   metavar='PASSWORD',
+                   help='The password for the SVN repository.'),
+            Option('--svn-prompt-password',
+                   dest='svn_prompt_password',
+                   default=False,
+                   action='store_true',
+                   help="Prompt for the user's svn password. This option "
+                        "overrides the password provided by the "
+                        "--svn-password option.",
+                   added_in='0.7.3'),
             Option('--svn-show-copies-as-adds',
                    dest='svn_show_copies_as_adds',
-                   metavar='y/n',
+                   metavar='y|n',
                    default=None,
-                   help='Treat copied or moved files as new files.'),
+                   help='Treat copied or moved files as new files.'
+                        '\n'
+                        'This is only supported in Subversion 1.7+.',
+                   added_in='0.5.2'),
             Option('--svn-changelist',
                    dest='svn_changelist',
                    default=None,
                    metavar='ID',
                    help='Generates the diff for review based on a '
-                        'local changelist. [DEPRECATED]'),
+                        'local changelist.',
+                   deprecated_in='0.6'),
+        ]
+    )
+
+    tfs_options = OptionGroup(
+        name='TFS Options',
+        description='Team Foundation Server specific options for '
+                    'communicating with the TFS server.',
+        option_list=[
+            Option('--tfs-login',
+                   dest='tfs_login',
+                   default=None,
+                   metavar='TFS_LOGIN',
+                   help='Logs in to TFS as a specific user (ie.'
+                        'user@domain,password). Visit https://msdn.microsoft.'
+                        'com/en-us/library/hh190725.aspx to learn about '
+                        'saving credentials for reuse.'),
         ]
     )
 
@@ -313,9 +494,11 @@ class Command(object):
 
     def create_parser(self, config, argv=[]):
         """Create and return the argument parser for this command."""
-        parser = argparse.ArgumentParser(prog=RB_MAIN,
-                                         usage=self.usage(),
-                                         add_help=False)
+        parser = argparse.ArgumentParser(
+            prog=RB_MAIN,
+            usage=self.usage(),
+            add_help=False,
+            formatter_class=SmartHelpFormatter)
 
         for option in self.option_list:
             option.add_to(parser, config, argv)
@@ -325,6 +508,15 @@ class Command(object):
 
         return parser
 
+    def post_process_options(self):
+        if self.options.disable_ssl_verification:
+            try:
+                import ssl
+                ssl._create_unverified_context()
+            except:
+                raise CommandError('The --disable-ssl-verification flag is '
+                                   'only available with Python 2.7.9+')
+
     def usage(self):
         """Return a usage string for the command."""
         usage = '%%(prog)s %s [options] %s' % (self.name, self.args)
@@ -402,19 +594,23 @@ class Command(object):
         minargs = len(argspec[0]) - 1
         maxargs = minargs
 
+        # Arguments that have a default value are considered optional.
+        if argspec[3] is not None:
+            minargs -= len(argspec[3])
+
         if argspec[1] is not None:
             maxargs = None
 
         if len(args) < minargs or (maxargs is not None and
                                    len(args) > maxargs):
-            parser.error("Invalid number of arguments provided")
+            parser.error('Invalid number of arguments provided')
             sys.exit(1)
 
         self.init_logging()
 
         try:
             exit_code = self.main(*args) or 0
-        except CommandError, e:
+        except CommandError as e:
             if isinstance(e, ParseError):
                 parser.error(e)
             elif self.options.debug:
@@ -422,9 +618,9 @@ class Command(object):
 
             logging.error(e)
             exit_code = 1
-        except CommandExit, e:
+        except CommandExit as e:
             exit_code = e.exit_code
-        except Exception, e:
+        except Exception as e:
             # If debugging is on, we'll let python spit out the
             # stack trace and report the exception, otherwise
             # we'll suppress the trace and print the exception
@@ -440,14 +636,13 @@ class Command(object):
 
     def initialize_scm_tool(self, client_name=None):
         """Initialize the SCM tool for the current working directory."""
-        repository_info, tool = scan_usable_client(self.options,
+        repository_info, tool = scan_usable_client(self.config,
+                                                   self.options,
                                                    client_name=client_name)
-        tool.user_config = self.config
-        tool.configs = [self.config]
 
         try:
             tool.check_options()
-        except OptionsCheckError, e:
+        except OptionsCheckError as e:
             sys.stderr.write('%s\n' % e)
             sys.exit(1)
 
@@ -469,8 +664,8 @@ class Command(object):
             server_url = tool.scan_for_server(repository_info)
 
         if not server_url:
-            print ("Unable to find a Review Board server "
-                   "for this source code tree.")
+            print('Unable to find a Review Board server for this source code '
+                  'tree.')
             sys.exit(1)
 
         return server_url
@@ -483,24 +678,23 @@ class Command(object):
         username and password. This is used as a callback in the
         API when the user requires authorization.
         """
-        if getattr(self.options, 'diff_filename', None) == '-':
-            die('HTTP authentication is required, but cannot be '
-                'used with --diff-filename=-')
-
         if username is None or password is None:
-            print
-            print "==> HTTP Authentication Required"
-            print 'Enter authorization information for "%s" at %s' % \
-                (realm, urlparse(uri)[1])
+            if getattr(self.options, 'diff_filename', None) == '-':
+                die('HTTP authentication is required, but cannot be '
+                    'used with --diff-filename=-')
 
-            # getpass will write its prompt to stderr but raw_input
+            print()
+            print('Please log in to the Review Board server at %s.' %
+                  urlparse(uri)[1])
+
+            # getpass will write its prompt to stderr but input
             # writes to stdout. See bug 2831.
             if username is None:
                 sys.stderr.write('Username: ')
-                username = raw_input()
+                username = input()
 
             if password is None:
-                password = getpass.getpass('Password: ')
+                password = getpass.getpass(b'Password: ')
 
         return username, password
 
@@ -515,23 +709,24 @@ class Command(object):
             die('A two-factor authentication token is required, but cannot '
                 'be used with --diff-filename=-')
 
-        print
-        print '==> Two-factor authentication token required'
+        print()
+        print('Please enter your two-factor authentication token for Review '
+              'Board.')
 
         if token_method == 'sms':
-            print ('You should be getting a text message with '
-                   'an authentication token.')
-            print 'Enter the token below.'
+            print('You should be getting a text message with '
+                  'an authentication token.')
+            print('Enter the token below.')
         elif token_method == 'call':
-            print ('You should be getting an automated phone call with '
-                   'an authentication token.')
-            print 'Enter the token below.'
+            print('You should be getting an automated phone call with '
+                  'an authentication token.')
+            print('Enter the token below.')
         elif token_method == 'generator':
-            print 'Enter the token shown on your token generator app below.'
+            print('Enter the token shown on your token generator app below.')
 
-        print
+        print()
 
-        return getpass.getpass('Token: ')
+        return getpass.getpass(b'Token: ')
 
     def _make_api_client(self, server_url):
         """Return an RBClient object for the server.
@@ -539,12 +734,19 @@ class Command(object):
         The RBClient will be instantiated with the proper arguments
         for talking to the provided Review Board server url.
         """
-        return RBClient(server_url,
-                        username=self.options.username,
-                        password=self.options.password,
-                        auth_callback=self.credentials_prompt,
-                        otp_token_callback=self.otp_token_prompt,
-                        disable_proxy=not self.options.enable_proxy)
+        return RBClient(
+            server_url,
+            username=self.options.username,
+            password=self.options.password,
+            api_token=self.options.api_token,
+            auth_callback=self.credentials_prompt,
+            otp_token_callback=self.otp_token_prompt,
+            disable_proxy=not self.options.enable_proxy,
+            verify_ssl=not self.options.disable_ssl_verification,
+            allow_caching=not self.options.disable_cache,
+            cache_location=self.options.cache_location,
+            in_memory_cache=self.options.in_memory_cache,
+            save_cookies=self.options.save_cookies)
 
     def get_api(self, server_url):
         """Returns an RBClient instance and the associated root resource.
@@ -552,15 +754,18 @@ class Command(object):
         Commands should use this method to gain access to the API,
         instead of instantianting their own client.
         """
+        if not urlparse(server_url).scheme:
+            server_url = '%s%s' % ('http://', server_url)
+
         api_client = self._make_api_client(server_url)
 
         try:
             api_root = api_client.get_root()
-        except ServerInterfaceError, e:
-            raise CommandError("Could not reach the Review Board "
-                               "server at %s" % server_url)
-        except APIError, e:
-            raise CommandError("Unexpected API Error: %s" % e)
+        except ServerInterfaceError as e:
+            raise CommandError('Could not reach the Review Board '
+                               'server at %s: %s' % (server_url, e))
+        except APIError as e:
+            raise CommandError('Unexpected API Error: %s' % e)
 
         return api_client, api_root
 
