@@ -1,23 +1,26 @@
+from __future__ import print_function, unicode_literals
+
 import logging
 import os
 import re
 import sys
 
 from rbtools.api.errors import APIError
-from rbtools.clients.errors import InvalidRevisionSpecError
 from rbtools.commands import Command, CommandError, Option, OptionGroup
+from rbtools.utils.commands import (AlreadyStampedError,
+                                    get_review_request,
+                                    stamp_commit_with_review_url)
 from rbtools.utils.console import confirm
-from rbtools.utils.match_score import Score
-from rbtools.utils.repository import get_repository_id
-from rbtools.utils.users import get_user
-
+from rbtools.utils.review_request import (get_draft_or_current_value,
+                                          get_revisions,
+                                          guess_existing_review_request)
 
 class Post(Command):
     """Create and update review requests."""
-    name = "post"
-    author = "The Review Board Project"
-    description = "Uploads diffs to create and update review requests."
-    args = "[revisions]"
+    name = 'post'
+    author = 'The Review Board Project'
+    description = 'Uploads diffs to create and update review requests.'
+    args = '[revisions]'
 
     GUESS_AUTO = 'auto'
     GUESS_YES = 'yes'
@@ -33,24 +36,29 @@ class Post(Command):
                         'review request gets posted and how, and what '
                         'happens after it is posted.',
             option_list=[
+                Option('-u', '--update',
+                       dest='update',
+                       action='store_true',
+                       default=False,
+                       help='Automatically determines the existing review '
+                            'request to update.',
+                       added_in='0.5.3'),
                 Option('-r', '--review-request-id',
                        dest='rid',
                        metavar='ID',
                        default=None,
                        help='Specifies the existing review request ID to '
                             'update.'),
-                Option('-u', '--update',
-                       dest='update',
-                       action='store_true',
-                       default=False,
-                       help='Automatically determines the existing review '
-                            'request to update.'),
                 Option('-p', '--publish',
                        dest='publish',
                        action='store_true',
                        default=False,
                        config_key='PUBLISH',
-                       help='Immediately publishes the review request after '
+                       help='Publishes the review request immediately after '
+                            'posting.'
+                            '\n'
+                            'All required fields must already be filled in '
+                            'on the review request or must be provided when '
                             'posting.'),
                 Option('-o', '--open',
                        dest='open_browser',
@@ -59,13 +67,27 @@ class Post(Command):
                        default=False,
                        help='Opens a web browser to the review request '
                             'after posting.'),
+                Option('-s', '--stamp',
+                       dest='stamp_when_posting',
+                       action='store_true',
+                       config_key='STAMP_WHEN_POSTING',
+                       default=False,
+                       help='Stamps the commit message with the review '
+                            'request URL while posting the review.',
+                       added_in='0.7.3'),
                 Option('--submit-as',
                        dest='submit_as',
                        metavar='USERNAME',
                        config_key='SUBMIT_AS',
                        default=None,
-                       help='The user name to use as the author of the '
-                            'review request, instead of the logged in user.'),
+                       help='The username to use as the author of the '
+                            'review request, instead of the logged in user.',
+                       extended_help=(
+                           "This is useful when used in a repository's "
+                           "post-commit script to update or create review "
+                           "requests. See :ref:`automating-rbt-post` for "
+                           "more information on this use case."
+                       )),
                 Option('--change-only',
                        dest='change_only',
                        action='store_true',
@@ -97,8 +119,13 @@ class Post(Command):
                        default=GUESS_AUTO,
                        const=GUESS_YES,
                        choices=GUESS_CHOICES,
-                       help='Short-hand for --guess-summary '
-                            '--guess-description.'),
+                       help='Equivalent to setting both --guess-summary '
+                            'and --guess-description.',
+                       extended_help=(
+                           'This can optionally take a value to control the '
+                           'guessing behavior. See :ref:`guessing-behavior` '
+                           'for more information.'
+                       )),
                 Option('--guess-summary',
                        dest='guess_summary',
                        action='store',
@@ -108,7 +135,12 @@ class Post(Command):
                        const=GUESS_YES,
                        choices=GUESS_CHOICES,
                        help='Generates the Summary field based on the '
-                            'commit messages (Bazaar/Git/Mercurial only).'),
+                            'commit messages (Bazaar/Git/Mercurial only).',
+                       extended_help=(
+                           'This can optionally take a value to control the '
+                           'guessing behavior. See :ref:`guessing-behavior` '
+                           'for more information.'
+                       )),
                 Option('--guess-description',
                        dest='guess_description',
                        action='store',
@@ -118,18 +150,26 @@ class Post(Command):
                        const=GUESS_YES,
                        choices=GUESS_CHOICES,
                        help='Generates the Description field based on the '
-                            'commit messages (Bazaar/Git/Mercurial only).'),
+                            'commit messages (Bazaar/Git/Mercurial only).',
+                       extended_help=(
+                           'This can optionally take a value to control the '
+                           'guessing behavior. See :ref:`guessing-behavior` '
+                           'for more information.'
+                       )),
                 Option('--change-description',
                        default=None,
+                       metavar='TEXT',
                        help='A description of what changed in this update '
                             'of the review request. This is ignored for new '
                             'review requests.'),
                 Option('--summary',
                        dest='summary',
+                       metavar='TEXT',
                        default=None,
                        help='The new contents for the Summary field.'),
                 Option('--description',
                        dest='description',
+                       metavar='TEXT',
                        default=None,
                        help='The new contents for the Description field.'),
                 Option('--description-file',
@@ -140,6 +180,7 @@ class Post(Command):
                             'Description field.'),
                 Option('--testing-done',
                        dest='testing_done',
+                       metavar='TEXT',
                        default=None,
                        help='The new contents for the Testing Done field.'),
                 Option('--testing-done-file',
@@ -151,20 +192,26 @@ class Post(Command):
                 Option('--branch',
                        dest='branch',
                        config_key='BRANCH',
+                       metavar='BRANCH',
                        default=None,
-                       help='The branch the change will be committed on.'),
+                       help='The branch the change will be committed on or '
+                            'affects. This is a free-form field and does not '
+                            'control any behavior.'),
                 Option('--bugs-closed',
                        dest='bugs_closed',
+                       metavar='BUG_ID[,...]',
                        default=None,
                        help='The comma-separated list of bug IDs closed.'),
                 Option('--target-groups',
                        dest='target_groups',
                        config_key='TARGET_GROUPS',
+                       metavar='NAME[,...]',
                        default=None,
                        help='The names of the groups that should perform the '
                             'review.'),
                 Option('--target-people',
                        dest='target_people',
+                       metavar='USERNAME[,...]',
                        config_key='TARGET_PEOPLE',
                        default=None,
                        help='The usernames of the people who should perform '
@@ -172,24 +219,32 @@ class Post(Command):
                 Option('--depends-on',
                        dest='depends_on',
                        config_key='DEPENDS_ON',
+                       metavar='ID[,...]',
                        default=None,
-                       help='The new contents for the Depends On field.'),
+                       help='A comma-separated list of review request IDs '
+                            'that this review request will depend on.',
+                       added_in='0.6.1'),
                 Option('--markdown',
                        dest='markdown',
                        action='store_true',
                        config_key='MARKDOWN',
                        default=False,
                        help='Specifies if the summary and description should '
-                            'be interpreted as Markdown-formatted text '
-                            '(Review Board 2.0+ only).'),
+                            'be interpreted as Markdown-formatted text.'
+                            '\n'
+                            'This is only supported in Review Board 2.0+.',
+                       added_in='0.6'),
             ]
         ),
         Command.diff_options,
         Command.perforce_options,
         Command.subversion_options,
+        Command.tfs_options,
     ]
 
     def post_process_options(self):
+        super(Post, self).post_process_options()
+
         # -g implies --guess-summary and --guess-description
         if self.options.guess_fields:
             self.options.guess_fields = self.normalize_guess_value(
@@ -213,34 +268,32 @@ class Post(Command):
 
         # Only one of --description and --description-file can be used
         if self.options.description and self.options.description_file:
-            raise CommandError("The --description and --description-file "
-                               "options are mutually exclusive.\n")
+            raise CommandError('The --description and --description-file '
+                               'options are mutually exclusive.')
 
         # If --description-file is used, read that file
         if self.options.description_file:
             if os.path.exists(self.options.description_file):
-                fp = open(self.options.description_file, "r")
-                self.options.description = fp.read()
-                fp.close()
+                with open(self.options.description_file, 'r') as fp:
+                    self.options.description = fp.read()
             else:
                 raise CommandError(
-                    "The description file %s does not exist.\n" %
-                    self.options.description_file)
+                    'The description file %s does not exist.'
+                    % self.options.description_file)
 
         # Only one of --testing-done and --testing-done-file can be used
         if self.options.testing_done and self.options.testing_file:
-            raise CommandError("The --testing-done and --testing-done-file "
-                               "options are mutually exclusive.\n")
+            raise CommandError('The --testing-done and --testing-done-file '
+                               'options are mutually exclusive.')
 
         # If --testing-done-file is used, read that file
         if self.options.testing_file:
             if os.path.exists(self.options.testing_file):
-                fp = open(self.options.testing_file, "r")
-                self.options.testing_done = fp.read()
-                fp.close()
+                with open(self.options.testing_file, 'r') as fp:
+                    self.options.testing_done = fp.read()
             else:
-                raise CommandError("The testing file %s does not exist.\n" %
-                                   self.options.testing_file)
+                raise CommandError('The testing file %s does not exist.'
+                                   % self.options.testing_file)
 
         # If we have an explicitly specified summary, override
         # --guess-summary
@@ -258,6 +311,12 @@ class Post(Command):
             self.options.guess_description = self.normalize_guess_value(
                 self.options.guess_description, '--guess-description')
 
+        # If the --diff-filename argument is used, we can't do automatic
+        # updating.
+        if self.options.diff_filename and self.options.update:
+            raise CommandError('The --update option cannot be used when '
+                               'using --diff-filename.')
+
         # If we have an explicitly specified review request ID, override
         # --update
         if self.options.rid and self.options.update:
@@ -281,7 +340,8 @@ class Post(Command):
         with those one the server, and return the first match.
         """
         if isinstance(repository_info.path, list):
-            repositories = api_root.get_repositories()
+            repositories = api_root.get_repositories(only_fields='path',
+                                                     only_links='')
 
             try:
                 while True:
@@ -314,143 +374,11 @@ class Post(Command):
 
         return repository_info.path
 
-    def get_draft_or_current_value(self, field_name, review_request):
-        """Returns the draft or current field value from a review request.
-
-        If a draft exists for the supplied review request, return the draft's
-        field value for the supplied field name, otherwise return the review
-        request's field value for the supplied field name.
-        """
-        if review_request.draft:
-            fields = review_request.draft[0]
-        else:
-            fields = review_request
-
-        return fields[field_name]
-
-    def get_possible_matches(self, review_requests, summary, description,
-                             limit=5):
-        """Returns a sorted list of tuples of score and review request.
-
-        Each review request is given a score based on the summary and
-        description provided. The result is a sorted list of tuples containing
-        the score and the corresponding review request, sorted by the highest
-        scoring review request first.
-        """
-        candidates = []
-
-        # Get all potential matches.
-        try:
-            while True:
-                for review_request in review_requests:
-                    summary_pair = (
-                        self.get_draft_or_current_value(
-                            'summary', review_request),
-                        summary)
-                    description_pair = (
-                        self.get_draft_or_current_value(
-                            'description', review_request),
-                        description)
-                    score = Score.get_match(summary_pair, description_pair)
-                    candidates.append((score, review_request))
-
-                review_requests = review_requests.get_next()
-        except StopIteration:
-            pass
-
-        # Sort by summary and description on descending rank.
-        sorted_candidates = sorted(
-            candidates,
-            key=lambda m: (m[0].summary_score, m[0].description_score),
-            reverse=True
-        )
-
-        return sorted_candidates[:limit]
-
-    def num_exact_matches(self, possible_matches):
-        """Returns the number of exact matches in the possible match list."""
-        count = 0
-
-        for score, request in possible_matches:
-            if score.is_exact_match():
-                count += 1
-
-        return count
-
-    def guess_existing_review_request_id(self, repository_info, api_root,
-                                         api_client):
-        """Try to guess the existing review request ID if it is available.
-
-        The existing review request is guessed by comparing the existing
-        summary and description to the current post's summary and description,
-        respectively. The current post's summary and description are guessed if
-        they are not provided.
-
-        If the summary and description exactly match those of an existing
-        review request, the ID for which is immediately returned. Otherwise,
-        the user is prompted to select from a list of potential matches,
-        sorted by the highest ranked match first.
-        """
-        user = get_user(api_client, api_root, auth_required=True)
-        repository_id = get_repository_id(
-            repository_info, api_root, self.options.repository_name)
-
-        try:
-            # Get only pending requests by the current user for this
-            # repository.
-            review_requests = api_root.get_review_requests(
-                repository=repository_id, from_user=user.username,
-                status='pending', expand='draft')
-
-            if not review_requests:
-                raise CommandError('No existing review requests to update for '
-                                   'user %s.'
-                                   % user.username)
-        except APIError, e:
-            raise CommandError('Error getting review requests for user '
-                               '%s: %s' % (user.username, e))
-
-        summary = self.options.summary
-        description = self.options.description
-
-        if not summary or not description:
-            try:
-                commit_message = self.get_commit_message()
-
-                if commit_message:
-                    if not summary:
-                        summary = commit_message['summary']
-
-                    if not description:
-                        description = commit_message['description']
-            except NotImplementedError:
-                raise CommandError('--summary and --description are required.')
-
-        possible_matches = self.get_possible_matches(review_requests, summary,
-                                                     description)
-        exact_match_count = self.num_exact_matches(possible_matches)
-
-        for score, review_request in possible_matches:
-            # If the score is the only exact match, return the review request
-            # ID without confirmation, otherwise prompt.
-            if score.is_exact_match() and exact_match_count == 1:
-                return review_request.id
-            else:
-                question = ("Update Review Request #%s: '%s'? "
-                            % (review_request.id,
-                               self.get_draft_or_current_value(
-                                   'summary', review_request)))
-
-                if confirm(question):
-                    return review_request.id
-
-        return None
-
-    def post_request(self, repository_info, server_url, api_root,
+    def post_request(self, repository_info, repository, server_url, api_root,
                      review_request_id=None, changenum=None, diff_content=None,
                      parent_diff_content=None, commit_id=None,
                      base_commit_id=None,
-                     submit_as=None, retries=3):
+                     submit_as=None, retries=3, base_dir=None):
         """Creates or updates a review request, and uploads a diff.
 
         On success the review request id and url are returned.
@@ -460,59 +388,59 @@ class Post(Command):
                                                   'commit_ids')
 
         if review_request_id:
-            # Retrieve the review request corresponding to the provided id.
-            try:
-                review_request = api_root.get_review_request(
-                    review_request_id=review_request_id)
-            except APIError, e:
-                raise CommandError("Error getting review request %s: %s"
-                                   % (review_request_id, e))
+            review_request = get_review_request(
+                review_request_id, api_root,
+                only_fields='absolute_url,bugs_closed,id,status',
+                only_links='diffs,draft')
 
             if review_request.status == 'submitted':
                 raise CommandError(
-                    "Review request %s is marked as %s. In order to update "
-                    "it, please reopen the review request and try again."
+                    'Review request %s is marked as %s. In order to update '
+                    'it, please reopen the review request and try again.'
                     % (review_request_id, review_request.status))
         else:
             # No review_request_id, so we will create a new review request.
             try:
-                repository = (
-                    self.options.repository_name or
-                    self.options.repository_url or
-                    self.get_repository_path(repository_info, api_root))
+                # Until we are Python 2.7+ only, the keys in request_data have
+                # to be bytes. See bug 3753 for details.
                 request_data = {
-                    'repository': repository
+                    b'repository': repository
                 }
 
                 if changenum:
-                    request_data['changenum'] = changenum
+                    request_data[b'changenum'] = changenum
                 elif commit_id and supports_posting_commit_ids:
-                    request_data['commit_id'] = commit_id
+                    request_data[b'commit_id'] = commit_id
 
                 if submit_as:
-                    request_data['submit_as'] = submit_as
+                    request_data[b'submit_as'] = submit_as
 
-                review_request = api_root.get_review_requests().create(
-                    **request_data)
-            except APIError, e:
-                if e.error_code == 204 and changenum:  # Change number in use.
+                review_requests = api_root.get_review_requests(
+                    only_fields='',
+                    only_links='create')
+                review_request = review_requests.create(**request_data)
+            except APIError as e:
+                if e.error_code == 204 and changenum:
+                    # The change number is already in use. Get the review
+                    # request for that change and update it instead.
                     rid = e.rsp['review_request']['id']
                     review_request = api_root.get_review_request(
-                        review_request_id=rid)
+                        review_request_id=rid,
+                        only_fields='absolute_url,bugs_closed,id,status',
+                        only_links='diffs,draft')
 
                     if not self.options.diff_only:
                         review_request = review_request.update(
                             changenum=changenum)
                 else:
-                    raise CommandError("Error creating review request: %s" % e)
+                    raise CommandError('Error creating review request: %s' % e)
 
         if (not repository_info.supports_changesets or
             not self.options.change_only):
             try:
                 diff_kwargs = {
                     'parent_diff': parent_diff_content,
-                    'base_dir': (self.options.basedir or
-                                 repository_info.base_path),
+                    'base_dir': base_dir,
                 }
 
                 if (base_commit_id and
@@ -523,9 +451,9 @@ class Post(Command):
                     # the diff.
                     diff_kwargs['base_commit_id'] = base_commit_id
 
-                review_request.get_diffs().upload_diff(diff_content,
-                                                       **diff_kwargs)
-            except APIError, e:
+                review_request.get_diffs(only_fields='').upload_diff(
+                    diff_content, **diff_kwargs)
+            except APIError as e:
                 error_msg = [
                     u'Error uploading diff\n\n',
                 ]
@@ -551,9 +479,41 @@ class Post(Command):
                 raise CommandError(u'\n'.join(error_msg))
 
         try:
-            draft = review_request.get_draft()
-        except APIError, e:
-            raise CommandError("Error retrieving review request draft: %s" % e)
+            draft = review_request.get_draft(only_fields='commit_id')
+        except APIError as e:
+            raise CommandError('Error retrieving review request draft: %s' % e)
+
+        # Stamp the commit message with the review request URL before posting
+        # the review, so that we can use the stamped commit message when
+        # guessing the description. This enables the stamped message to be
+        # present on the review if the user has chosen to publish immediately
+        # upon posting.
+        if self.options.stamp_when_posting:
+            if not self.tool.can_amend_commit:
+                print('Cannot stamp review URL onto the commit message; '
+                      'stamping is not supported with %s.' % self.tool.name)
+
+            else:
+                try:
+                    stamp_commit_with_review_url(self.revisions,
+                                                 review_request.absolute_url,
+                                                 self.tool)
+                    print('Stamped review URL onto the commit message.')
+                except AlreadyStampedError:
+                    print('Commit message has already been stamped')
+                except Exception as e:
+                    logging.debug('Caught exception while stamping the '
+                                  'commit message. Proceeding to post '
+                                  'without stamping.', exc_info=True)
+                    print('Could not stamp review URL onto the commit '
+                          'message.')
+
+        # If the user has requested to guess the summary or description,
+        # get the commit message and override the summary and description
+        # options. The guessing takes place after stamping so that the
+        # guessed description matches the commit when rbt exits.
+        if not self.options.diff_filename:
+            self.check_guess_fields()
 
         # Update the review request draft fields based on options set
         # by the user, or configuration.
@@ -576,10 +536,10 @@ class Post(Command):
 
         if self.options.bugs_closed:
             # Append to the existing list of bugs.
-            self.options.bugs_closed = self.options.bugs_closed.strip(", ")
-            bug_set = (set(re.split("[, ]+", self.options.bugs_closed)) |
+            self.options.bugs_closed = self.options.bugs_closed.strip(', ')
+            bug_set = (set(re.split('[, ]+', self.options.bugs_closed)) |
                        set(review_request.bugs_closed))
-            self.options.bugs_closed = ",".join(bug_set)
+            self.options.bugs_closed = ','.join(bug_set)
             update_fields['bugs_closed'] = self.options.bugs_closed
 
         if self.options.description:
@@ -608,32 +568,16 @@ class Post(Command):
         if update_fields:
             try:
                 draft = draft.update(**update_fields)
-            except APIError, e:
+            except APIError as e:
                 raise CommandError(
-                    "Error updating review request draft: %s" % e)
+                    'Error updating review request draft: %s\n\n'
+                    'Your review request still exists, but the diff is not '
+                    'attached.\n\n'
+                    '%s\n'
+                    % (e, review_request.absolute_url))
 
         return review_request.id, review_request.absolute_url
 
-    def get_revisions(self):
-        """Returns the parsed revisions from the command line arguments.
-
-        These revisions are used for diff generation and commit message
-        extraction. They will be cached for future calls.
-        """
-        # Parse the provided revisions from the command line and generate
-        # a spec or set of specialized extra arguments that the SCMClient
-        # can use for diffing and commit lookups.
-        if not hasattr(self, '_revisions'):
-            try:
-                self._revisions = self.tool.parse_revision_spec(self.cmd_args)
-            except InvalidRevisionSpecError:
-                if not self.tool.supports_diff_extra_args:
-                    raise
-
-                self._revisions = None
-
-        return self._revisions
-
     def check_guess_fields(self):
         """Checks and handles field guesses for the review request.
 
@@ -657,9 +601,9 @@ class Post(Command):
             (self.options.guess_description == self.GUESS_AUTO and
              is_new_review_request))
 
-        if guess_summary or guess_description:
+        if self.revisions and (guess_summary or guess_description):
             try:
-                commit_message = self.get_commit_message()
+                commit_message = self.tool.get_commit_message(self.revisions)
 
                 if commit_message:
                     if guess_summary:
@@ -673,20 +617,13 @@ class Post(Command):
                 # so we can't provide the guessed versions.
                 pass
 
-    def get_commit_message(self):
-        """Returns the commit message for the parsed revisions.
+    def _ask_review_request_match(self, review_request):
+        question = ("Update Review Request #%s: '%s'? "
+                    % (review_request.id,
+                       get_draft_or_current_value(
+                           'summary', review_request)))
 
-        If the SCMClient supports getting a commit message, this will fetch
-        and store the message for future lookups.
-
-        This is used for guessing the summary and description fields, and
-        updating exising review requests using -u.
-        """
-        if not hasattr(self, '_commit_message'):
-            self._commit_message = \
-                self.tool.get_commit_message(self.get_revisions())
-
-        return self._commit_message
+        return confirm(question)
 
     def main(self, *args):
         """Create and update review requests."""
@@ -706,26 +643,41 @@ class Post(Command):
         api_client, api_root = self.get_api(server_url)
         self.setup_tool(self.tool, api_root=api_root)
 
+        if (self.options.exclude_patterns and
+            not self.tool.supports_diff_exclude_patterns):
+
+            raise CommandError(
+                'The %s backend does not support excluding files via the '
+                '-X/--exclude commandline options or the EXCLUDE_PATTERNS '
+                '.reviewboardrc option.' % self.tool.name)
+
+        # Check if repository info on reviewboard server match local ones.
+        repository_info = repository_info.find_server_repository_info(api_root)
+
         if self.options.diff_filename:
+            self.revisions = None
             parent_diff = None
             base_commit_id = None
             commit_id = None
 
             if self.options.diff_filename == '-':
-                diff = sys.stdin.read()
+                if hasattr(sys.stdin, 'buffer'):
+                    # Make sure we get bytes on Python 3.x
+                    diff = sys.stdin.buffer.read()
+                else:
+                    diff = sys.stdin.read()
             else:
                 try:
                     diff_path = os.path.join(origcwd,
                                              self.options.diff_filename)
-                    fp = open(diff_path, 'r')
-                    diff = fp.read()
-                    fp.close()
-                except IOError, e:
-                    raise CommandError("Unable to open diff filename: %s" % e)
+                    with open(diff_path, 'rb') as fp:
+                        diff = fp.read()
+                except IOError as e:
+                    raise CommandError('Unable to open diff filename: %s' % e)
         else:
-            revisions = self.get_revisions()
+            self.revisions = get_revisions(self.tool, self.cmd_args)
 
-            if revisions:
+            if self.revisions:
                 extra_args = None
             else:
                 extra_args = self.cmd_args
@@ -733,8 +685,9 @@ class Post(Command):
             # Generate a diff against the revisions or arguments, filtering
             # by the requested files if provided.
             diff_info = self.tool.diff(
-                revisions=revisions,
-                files=self.options.include_files or [],
+                revisions=self.revisions,
+                include_files=self.options.include_files or [],
+                exclude_patterns=self.options.exclude_patterns or [],
                 extra_args=extra_args)
 
             diff = diff_info['diff']
@@ -742,37 +695,96 @@ class Post(Command):
             base_commit_id = diff_info.get('base_commit_id')
             commit_id = diff_info.get('commit_id')
 
+        repository = (
+            self.options.repository_name or
+            self.options.repository_url or
+            self.get_repository_path(repository_info, api_root))
+
+        base_dir = self.options.basedir or repository_info.base_path
+
+        if repository is None:
+            raise CommandError('Could not find the repository on the Review '
+                               'Board server.')
+
         if len(diff) == 0:
             raise CommandError("There don't seem to be any diffs!")
 
+        # Validate the diffs to ensure that they can be parsed and that
+        # all referenced files can be found.
+        #
+        # Review Board 2.0.14+ (with the diffs.validation.base_commit_ids
+        # capability) is required to successfully validate against hosting
+        # services that need a base_commit_id. This is basically due to
+        # the limitations of a couple Git-specific hosting services
+        # (Beanstalk, Bitbucket, and Unfuddle).
+        #
+        # In order to validate, we need to either not be dealing with a
+        # base commit ID (--diff-filename), or be on a new enough version
+        # of Review Board, or be using a non-Git repository.
+        can_validate_base_commit_ids = \
+            self.tool.capabilities.has_capability('diffs', 'validation',
+                                                  'base_commit_ids')
+
+        if (not base_commit_id or
+            can_validate_base_commit_ids or
+            self.tool.name != 'Git'):
+            # We can safely validate this diff before posting it, but we
+            # need to ensure we only pass base_commit_id if the capability
+            # is set.
+            validate_kwargs = {}
+
+            if can_validate_base_commit_ids:
+                validate_kwargs['base_commit_id'] = base_commit_id
+
+            try:
+                diff_validator = api_root.get_diff_validation()
+                diff_validator.validate_diff(
+                    repository,
+                    diff,
+                    parent_diff=parent_diff,
+                    base_dir=base_dir,
+                    **validate_kwargs)
+            except APIError as e:
+                msg_prefix = ''
+
+                if e.error_code == 207:
+                    msg_prefix = '%s: ' % e.rsp['file']
+
+                raise CommandError('Error validating diff\n\n%s%s' %
+                                   (msg_prefix, e))
+            except AttributeError:
+                # The server doesn't have a diff validation resource. Post as
+                # normal.
+                pass
+
         if repository_info.supports_changesets and 'changenum' in diff_info:
             changenum = diff_info['changenum']
             commit_id = changenum
         else:
             changenum = None
 
-        if not self.options.diff_filename:
-            # If the user has requested to guess the summary or description,
-            # get the commit message and override the summary and description
-            # options.
-            self.check_guess_fields()
+        if self.options.update and self.revisions:
+            review_request = guess_existing_review_request(
+                repository_info, self.options.repository_name, api_root,
+                api_client, self.tool, self.revisions,
+                guess_summary=False, guess_description=False,
+                is_fuzzy_match_func=self._ask_review_request_match)
 
-        if self.options.update:
-            self.options.rid = self.guess_existing_review_request_id(
-                repository_info, api_root, api_client)
-
-            if not self.options.rid:
+            if not review_request or not review_request.id:
                 raise CommandError('Could not determine the existing review '
                                    'request to update.')
 
+            self.options.rid = review_request.id
+
         # If only certain files within a commit are being submitted for review,
-        # do not include the commit id. This prevents conflicts if mutliple
+        # do not include the commit id. This prevents conflicts if multiple
         # files from the same commit are posted for review separately.
-        if self.options.include_files:
+        if self.options.include_files or self.options.exclude_patterns:
             commit_id = None
 
         request_id, review_url = self.post_request(
             repository_info,
+            repository,
             server_url,
             api_root,
             self.options.rid,
@@ -781,14 +793,15 @@ class Post(Command):
             parent_diff_content=parent_diff,
             commit_id=commit_id,
             base_commit_id=base_commit_id,
-            submit_as=self.options.submit_as)
+            submit_as=self.options.submit_as,
+            base_dir=base_dir)
 
         diff_review_url = review_url + 'diff/'
 
-        print "Review request #%s posted." % request_id
-        print
-        print review_url
-        print diff_review_url
+        print('Review request #%s posted.' % request_id)
+        print()
+        print(review_url)
+        print(diff_review_url)
 
         # Load the review up in the browser if requested to.
         if self.options.open_browser:
